//@+leo-ver=4
//@+node:@file host/base.c
//@@language c
/**
 @file
 */

/**
 * @defgroup group_linux Luddite Linux Side Library
 * These routines are made available by lud-core to all luddite drivers
 * and linux side modules.
 * @{
 */

/* Copyright Philips Research Labs, Redhill, UK. 2001
 *
 * Prototype Software
 *
 * Contact: Tim Everett (tim.everett@philips.com)
 *
 */

/***** Includes *****/

//@<< lud-base #includes >>
//@+node:<< lud-base #includes >>
#ifdef CONFIG_SMP
#define __SMP__
#endif

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/page.h>
#include <asm/semaphore.h>
#include <asm/atomic.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>

#include <linux/module.h>

#include <linux/version.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/kmod.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#include <linux/device.h>
#include <linux/cdev.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18))
#include <linux/compiler.h>
#else
#define likely(x) (x)
#define unlikely(x) (x)
#endif /* LINUX_VERSION_CODE */

#include <stdarg.h>


#include "compat.h"
#include "lud_mmio.h"

#include "luddite-host.h"
#include "luddite/version.h"
#include "luddite/os.h"
#include "exported.h"

#include "core.h"
#include "procfs.h"
#include "node.h"
#include "helpers.h"
#include "channels.h"
//@-node:<< lud-base #includes >>
//@nl
//@<< lud-base declarations >>
//@+node:<< lud-base declarations >>

int lud_debug = 0;
int clear_target = 0;

struct LUD_DEV_STRUCT lud_devices[MAX_LUD_DEV];

static struct semaphore tms_sem ;

//@-node:<< lud-base declarations >>
//@nl
//@+others
//@+node:broadcast_devices
//@+at
//@nonl
// * broadcast current luddite Targets to all Targets running luddite * * not
// efficient, but for small numbers of Targets should be OK. * future
// improvements may be to only broadcast new data (add an extra * field to the
// lud_devices structure.
//@-at
//@@c

static void broadcast_devices(void)
{
	int i;
	for (i=0;i<MAX_LUD_DEV;i++) {
		if (lud_devices[i].luddev && (lud_devices[i].luddev->state == LUD_LUDDITE) ) {
			int j;
			for (j=0;j<MAX_LUD_DEV;j++) {
				LUD_DEV* t = lud_devices[j].luddev;
				if (t && (t->state == LUD_LUDDITE) ) {
					/* report details for j to i */
					if (lud_devices[i].system) {
						LUD_DEV_LINK* clone= lud_clone_link(lud_devices[i].system);
						lud_send_command(clone,
							LUD_SYS_ADD_TARGET,
							j,
							t->comvec,
							t->mmio_phy,t->mmio_offset);
						lud_free_link(clone);
					} else {
						printk("lud: spurious broadcast for %d",i);
					}
				}
			}
		}
	}
}
//@-node:broadcast_devices
//@+node:lud_class_release


#ifdef CONFIG_SYSFS
/*
 * set up bus type
 */

static void lud_class_release(struct class_device* cd)
{
}

struct class lud_class = {
	.name = "luddite",
	.owner = THIS_MODULE,
	.release = lud_class_release,
};

#endif
//@-node:lud_class_release
//@+node:Printk
/* ************************************************************************ *
 * Macros                                                                   *
 * ************************************************************************ */

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#define INC_LUD_MOD_USE_COUNT(tmlink_ponter) do {} while (0);
#define DEC_LUD_MOD_USE_COUNT(tmlink_ponter) do {} while (0);
#else
#define INC_LUD_MOD_USE_COUNT(tmlink_ponter) \
    __MOD_INC_USE_COUNT(lud_devices[(tmlink_ponter)->minor].luddev->driver_module)
#define DEC_LUD_MOD_USE_COUNT(tmlink_ponter) \
    __MOD_DEC_USE_COUNT(lud_devices[(tmlink_ponter)->minor].luddev->driver_module)
#endif

/*
 * printf(HERE_STRING_FORMAT, HERE_STRING_ARGS) results in a string of the
 * form "file.c:123 (some_func)".   This uses a GCC extension.  Other
 * compilers will return just "file.c:123".
 */
#ifdef __GNUC__
# define HERE_STRING_FORMAT "%s:%d (%s)"
# define HERE_STRING_ARGS   __FILE__, __LINE__, __PRETTY_FUNCTION__
#else
# define HERE_STRING_FORMAT "%s:%d"
# define HERE_STRING_ARGS   __FILE__, __LINE__
#endif

#undef KERN_DEBUG
#define KERN_DEBUG
#define DEB(n,x) do { if (lud_debug >= n) x; } while (0)


/* ************************************************************************ *
 * Function prototypes                                                      *
 * ************************************************************************ */

/**
 * @internal
 * Create a new channel with given identifier
 */

/**
 * @internal
 * The default handler for messages from the luddite system thread
 */
static void lud_system_handler(LUD_COMMAND*,LUD_DEV*);
/**
 * @internal
 * The default handler for messages from the luddite loader thread
 */
static void lud_loader_handler(LUD_COMMAND*,LUD_DEV*);


/* ************************************************************************ *
 * Variables                                                                *
 * ************************************************************************ */

#ifndef CONFIG_SYSFS
static devfs_handle_t luddite_devfs_dir;
#endif
#if 0
/**
 * Handy printk that just prints to the console.
 */
void Printk(char* format, ...)
{
	va_list args;
	char PK_buff[256];	/* be careful with restricted size on stack */

	va_start(args,format);
	vsnprintf(PK_buff,sizeof(PK_buff)-1,format,args);
	va_end(args);
	console_print(PK_buff);
}
#endif
//@-node:Printk
//@+node:lud_dev_release

/* ************************************************************************ *
 * Node creation/deletion                                                   *
 * ************************************************************************ */
#ifdef CONFIG_SYSFS
static void lud_dev_release(struct device* dev)
{
}
#endif
//@nonl
//@-node:lud_dev_release
//@+node:lud_new_dev
//@+at
//@nonl
// ************************************************************************ *
// * Link creation/deletion                          * *
// ************************************************************************
//@-at
//@@c

static LUD_DEV_LINK* lud_new_dev(int major,int minor,lud_handler_func_t h)
{
	LUD_DEV_LINK* tm = NULL;
	struct LUD_NODE_DEV* node ;

	major = LUD_NODE_FROM_HANDLE(major);

	if (unlikely((minor < 0) || (minor >= MAX_LUD_DEV)))
		return NULL;
	if (unlikely((major < 0) || (major >= LUD_MAX_NODES)))
		return NULL;

	node = lud_devices[minor].lud_nodes[major];
	if (node==NULL) {
		node = (struct LUD_NODE_DEV*)kmalloc(sizeof(struct LUD_NODE_DEV),GFP_KERNEL);
		if (unlikely(!node)) {
			return NULL ;
		}
		memset(node,0,sizeof(struct LUD_NODE_DEV));
		lud_devices[minor].lud_nodes[major] = node;
	}

	tm = &lud_devices[minor].lud_nodes[major]->dev;
	tm->major = major;
	tm->minor = minor;
	tm->dev = lud_devices[minor].luddev;

	atomic_set(&tm->freecounter, 1);
	init_waitqueue_head(&tm->q);

	lud_devices[minor].lud_nodes[major]->handler = h;
	return tm;
}
//@-node:lud_new_dev
//@+node:lud_new_link
//@+at
//@nonl
// * * Create a new link to the given module on the given TriMedia. * @param
// major The module ID of the target component. * @param minor The ID of the
// TriMedia where the target module is running. * @returns  A pointer to a new
// link.
//@-at
//@@c

LUD_DEV_LINK* lud_new_link(int major,int minor)
{
	LUD_DEV_LINK* tm ;

	major = LUD_NODE_FROM_HANDLE(major);

	if (unlikely((minor < 0) || (minor >= MAX_LUD_DEV)))
		return NULL;
	if (unlikely((major < 0) || (major >= LUD_MAX_NODES)))
		return NULL;

	tm = (LUD_DEV_LINK*)kmalloc(sizeof(LUD_DEV_LINK),GFP_KERNEL);
	if (unlikely(!tm)) {
		return NULL;
	}

	tm->major = major;
	tm->minor = minor;
	tm->dev = lud_devices[tm->minor].luddev;
	atomic_set(&tm->freecounter, 1);
	init_waitqueue_head(&tm->q);
	return tm;
}
//@-node:lud_new_link
//@+node:lud_clone_link
//@+at
//@nonl
// * * Make a copy of a link - to support asynchronous communication. * @param
// s The link to clone. * @returns A copy of the given link.
//@-at
//@@c

LUD_DEV_LINK* lud_clone_link(LUD_DEV_LINK* s)
{
	LUD_DEV_LINK* tm = (LUD_DEV_LINK*)kmalloc(sizeof(LUD_DEV_LINK),GFP_KERNEL);

	if (unlikely(!tm)) {
		return NULL;
	}
	tm->major = s->major;
	tm->minor = s->minor;
	tm->dev = s->dev;
	atomic_set(&tm->freecounter, 1);
	init_waitqueue_head(&tm->q);
	return tm;
}
//@-node:lud_clone_link
//@+node:lud_free_link
//@+at
//@nonl
// * * Free the linux resources associated with the given link. * @param tm
// The link.
//@-at
//@@c

void lud_free_link(LUD_DEV_LINK* tm)
{
	if ( ((char *)tm >= (char *)lud_devices) && ((char *)tm < sizeof(lud_devices) + (char *)lud_devices) )
	{
		printk(KERN_CRIT "lud-core detected a module bug: for lud_get_node_link(), use lud_free_node_link() not lud_free_link()\n");
		lud_free_node_link(tm);
		BUG();
		return;
	}

	if (atomic_dec_and_test(&tm->freecounter))
	{
		kfree(tm);
	}
}
//@-node:lud_free_link
//@+node:initstr
//@+at
//@nonl
// ************************************************************************ *
// * Other functions...                            * *
// ************************************************************************
//@-at
//@@c

void initstr(int j)
{
	int i;

	for (i=0;i<LUD_MAX_NODES;i++) {
		lud_devices[j].lud_nodes[i] = NULL;
	}

	for (i=0;i<LUDMOD_MAX_MODULES;i++) {
		lud_devices[j].lud_modules[i].name = NULL;
		lud_devices[j].lud_modules[i].opts = NULL;
		lud_devices[j].lud_modules[i].pri = 0;
	}

	lud_devices[j].system = lud_new_dev(LUD_SYSTEM_MAJOR,j,lud_system_handler);
	lud_devices[j].loader = lud_new_dev(LUD_LOADER_MAJOR,j,lud_loader_handler);
}
//@-node:initstr
//@+node:cleanstr

void cleanstr(int j)
{
	int i;

	/* Delete modules */
	for (i=0;i<LUDMOD_MAX_MODULES;i++) {
		if (lud_devices[j].lud_modules[i].name) {
			kfree(lud_devices[j].lud_modules[i].name);
			lud_devices[j].lud_modules[i].name = NULL;
		}
		if (lud_devices[j].lud_modules[i].opts) {
			kfree(lud_devices[j].lud_modules[i].opts);
			lud_devices[j].lud_modules[i].opts = NULL;
		}
	}
}
//@-node:cleanstr
//@+node:install_mod

void install_mod(int m,char* n,char* o,int p,int l)
{
	if (lud_devices[m].lud_modules[l].name) {
		kfree(lud_devices[m].lud_modules[l].name);
		lud_devices[m].lud_modules[l].name = NULL;
	}

	if (lud_devices[m].lud_modules[l].opts) {
		kfree(lud_devices[m].lud_modules[l].opts);
		lud_devices[m].lud_modules[l].opts = NULL;
	}

	lud_devices[m].lud_modules[l].name = n;
	lud_devices[m].lud_modules[l].opts = o;
	lud_devices[m].lud_modules[l].pri = p;
}
//@-node:install_mod
//@+node:lud_wakeup

void lud_wakeup(LUD_DEV_LINK* tm)
{
	if (unlikely(tm == NULL)) {
		printk(KERN_EMERG "bug: lud_wakeup: Null link\n");
		return ;
	}
	tm->status = 1;
	wake_up_interruptible(&tm->q);
}
//@-node:lud_wakeup
//@+node:lud_send_command
//@+at
//@nonl
// * * Send a command to the TriMedia. There is no reply. * * @param tm  The
// link to use. * @param com The command to send. * @param a0  The first
// argument to the command. * @param a1  The second argument to the command. *
// @param a2  The third argument to the command. * @param a3  The fourth
// argument to the command.
//@-at
//@@c

void lud_send_command(LUD_DEV_LINK* tm,int com,int a0,int a1,int a2, int a3)
{
	DEB(1,printk(KERN_DEBUG "send_command %x\n",com));
	if (unlikely(tm == NULL)) {
		printk(KERN_EMERG "lud_send_command: Null link\n");
	}
	else {
		if (lud_devices[tm->minor].luddev->send_command) {
			lud_devices[tm->minor].luddev->send_command(tm,com,a0,a1,a2,a3);
		}
	}
}
//@-node:lud_send_command
//@+node:lud_send_ack
//@+at
//@nonl
// * * Send an acknowledgement to the TriMedia. This is used * to send a reply
// to a recieved request, it should not * be used to initiate communication. *
// * @param tm  The link to use. * @param reply An opaque number passed by the
// TriMedia *       when it sent the request. * @param what The reply to send.
//@-at
//@@c

void lud_send_ack(LUD_DEV_LINK* tm,LUD_REPLY reply,int what)
{
	DEB(1,printk(KERN_DEBUG "send_ack %x:%p\n",reply.cpu,
		    (void *)token_to_addr(reply.pointer)));
	if (unlikely(tm == NULL)) {
		printk(KERN_EMERG "lud_send_ack: Null link\n");
	}
	else {
		if (lud_devices[tm->minor].luddev->send_ack) {
			lud_devices[tm->minor].luddev->send_ack(tm,reply,what);
		}
	}
}
//@-node:lud_send_ack
//@+node:lud_send_request
//@+at
//@nonl
// * * Send a request to the TriMedia, and wait for a reply. * Can be
// interrupted by a Linux signal - e.g. Ctrl-C or SIGKILL. * * @param tm  The
// link to use. Only one request may be in progress *       on a single
// LUD_DEV_LINK at any one time. You can use *       lud_clone_link() to have
// unlimited numbers of LUD_DEV_LINK *       structures and hence unlimited
// simultaneous requests. * @param com The command to send. * @param a0  The
// first argument to the command. * @param a1  The second argument to the
// command. * @param a2  The third argument to the command. * @param a3  The
// fourth argument to the command. * @return   The result from the TriMedia,
// or -1 if the process is *       interrupted (e.g. by Ctrl-C or SIGKILL).
//@-at
//@@c

int lud_send_request(LUD_DEV_LINK* tm,int com,int a0,int a1,int a2, int a3)
{
	DEB(1,printk(KERN_DEBUG "send_request %x\n",com));
	if (unlikely(tm == NULL)) {
		printk(KERN_EMERG "lud_send_request: Null link\n");
		return 0;
	}
	if (lud_devices[tm->minor].luddev->send_request) {
		return lud_devices[tm->minor].luddev->send_request(tm,com,a0,a1,a2,a3);
	}
	return 0;
}
//@-node:lud_send_request
//@+node:uninstall_mod
//@+at
//@nonl
// ************************************************************************ *
// * Other functions...                            * *
// ************************************************************************
//@-at
//@@c

int uninstall_mod(int m,char* n)
{
	int i;
	int major = 0;
	int status;

	if (unlikely((m < 0) || (m >= MAX_LUD_DEV))) {
		return -EINVAL ;
	}

	for (i=0;i<LUDMOD_MAX_MODULES;i++) {
		if (lud_devices[m].lud_modules[i].name && (strcmp(lud_devices[m].lud_modules[i].name,n))==0) {
			major = i;
			break;
		}
	}
	if (major == 0) {
		return -ENODEV;
	}

	{
		LUD_DEV_LINK *unload_link = lud_clone_link(lud_devices[m].loader);
		status = lud_send_request(unload_link,LUD_SYS_UNLOAD,major,0,0,0);
		lud_free_link(unload_link);
	}

	switch (status) {
	case 0:
	case 1:
		return -EINVAL;
	case LUD_LOAD_SYSERROR:
		return -EIO;
	case LUD_LOAD_LOGICERROR:
		return -EBUSY;
	case LUD_LOAD_FILEERROR:
		return -ENOEXEC;
	}

	if (lud_devices[m].lud_modules[major].name) {
		kfree(lud_devices[m].lud_modules[major].name);
		lud_devices[m].lud_modules[major].name = NULL;
	}
	if (lud_devices[m].lud_modules[major].opts) {
		kfree(lud_devices[m].lud_modules[major].opts);
		lud_devices[m].lud_modules[major].opts = NULL;
	}

	return 0;
}
//@-node:uninstall_mod
//@+node:lud_remap_addr
//@+at
//@nonl
// * * Map a Target address to a Linux kernel address. * @param m a pointer to
// the TriMedia device structure * @param addr the address as the TriMedia
// sees it * @returns the corresponding address in kernel space.
//@-at
//@@c

unsigned long lud_remap_addr(LUD_DEV* m,unsigned long addr)
{
	/*  For 64 bit hosts, we throw away the top 32 bit
            of sdram_phy */
        addr -= (m->sdram_phy & 0xffffffff);
	if (unlikely(addr > m->sdram_size))
		return 0;
	return addr + m->sdram_log;
}
//@-node:lud_remap_addr
//@+node:lud_loader_handler
//@+at
//@nonl
// ************************************************************************ *
// * System and loader thread handlers                    * *
// ************************************************************************
//@-at
//@@c

static unsigned char* lud_dummy_file = NULL;
static int lud_dummy_offset;
static int lud_dummy_size;

static void lud_loader_handler(LUD_COMMAND* tc,LUD_DEV* m) {
}
//@-node:lud_loader_handler
//@+node:lud_system_handler

#define SCRATCH_SIZE 256 /* Luddite printk buffer size, careful it's not too big */

static void lud_system_handler(LUD_COMMAND* tc,LUD_DEV* m) {
	DEB(1,printk(KERN_DEBUG "system called %x\n",tc->command));
	switch (tc->command) {
	case LUD_SYS_LUDDITE_STARTED:
		m->state = LUD_LUDDITE;
		broadcast_devices();
		wake_up_interruptible(&m->start_queue);
		break;
	case LUD_SYS_NOOP:
		break;
	case LUD_SYS_PAGE_SIZE:
		lud_send_ack(lud_devices[m->minor].system,tc->reply,PAGE_SIZE);
		break;
	case LUD_SYS_PRINTK:
		{
			unsigned int addr = (unsigned long)tc->arg0;
			unsigned char* p = (unsigned char*)lud_remap_addr(m,addr);
			if (p==NULL)
				printk(KERN_EMERG "printk bad pointer from Target %d (0x%x)\n", m->minor,addr);
			else {
				unsigned char printk_buffer[SCRATCH_SIZE];

				memcpy_fromio(printk_buffer, p, sizeof(printk_buffer));
				printk("%s",printk_buffer);
			}
			lud_send_ack(lud_devices[m->minor].system,tc->reply,0);
		}
		break;
	case LUD_SYS_NEW_NODE:
		{
			unsigned int addr = (unsigned long)tc->arg1;
			unsigned char* p = (unsigned char*)lud_remap_addr(m,addr);

			if (unlikely(p == NULL)) {
				printk(KERN_EMERG "new_node bad pointer from Target %d (0x%x)\n", m->minor, addr);
				return ;
			}
			else {
				char name[LUD_MAX_NODE_NAME_LENGTH];

				strncpy_fromio(name, p, sizeof(name)-1);
				new_node(m->minor,tc->arg0,name,tc->arg2);
			}
			lud_send_ack(lud_devices[m->minor].system,tc->reply,0);
			break;
		}
	case LUD_SYS_DEL_NODE:
		del_node(m->minor,tc->arg0);
		break;
	case LUD_SYS_READ:
		{
			unsigned int addr = (unsigned long)tc->arg0;
			unsigned char* p = (unsigned char*)lud_remap_addr(m,addr);
			int Left = (lud_dummy_size - lud_dummy_offset);
			int ToCopy = (Left==0)?-1:((Left>=tc->arg1)?tc->arg1:Left);
			if (unlikely(p == NULL)) {
				printk(KERN_EMERG "PK:bad pointer address from Target %d(0x%x)\n",m->minor,addr);
				return ;
			}
			if (ToCopy>0) {
				memcpy_toio(p,&lud_dummy_file[lud_dummy_offset],ToCopy);
				lud_dummy_offset += ToCopy;
			}
			lud_send_ack(lud_devices[m->minor].system,tc->reply,ToCopy);
		}
		break;
	case LUD_SYS_CLOSE:
		if (lud_dummy_file) {
			kfree(lud_dummy_file);
			lud_dummy_file = 0;
		}
		break;
	case LUD_SYS_XFER_BUFFER:
		/* Need to free an Xfer.  Note that the main LUD_SYS_XFER_BUFFER
		 * handling code is in the ISR, for speed.  We only get here
		 * if the LUD_CHANNEL structure needs to be free()d.
		 */
		{
			LUD_CHANNEL * tmc = (LUD_CHANNEL *)token_to_addr(tc->arg3);
			if (unlikely(tmc == NULL)) {
				printk(KERN_EMERG "xfer_buffer bad pointer from Target %d (0x%x)\n", m->minor, tc->arg3);
			}
			else {
				token_free(tc->arg3);
				lud_free_link(tmc->luddev);
				kfree(tmc);
			}
		}
		break;
	case LUD_SYS_REQUEST_ACK:
		/* Need to free a LUD_DEV_LINK.  Note that the main LUD_SYS_REQUEST_ACK
		 * handling code is in the ISR, for speed.  We only get here
		 * if the LUD_DEV_LINK structure needs to be free()d.
		 */
		{
			LUD_DEV_LINK* tm = (LUD_DEV_LINK*)token_to_addr(tc->reply.pointer);
			if (unlikely(tm == NULL)) {
				printk(KERN_EMERG "request_ack bad pointer from Target %d (0x%x)\n", m->minor, tc->reply.pointer);
			}
			else {
				token_free(tc->reply.pointer);
				// 2005-07-19 Wouter Batelaan (WB) temporary patch - kfree(tm);
			}
		}
		break;
	}
}
//@-node:lud_system_handler
//@+node:lud_q_handler
//@+at
//@nonl
// ************************************************************************ *
// * Interrupt handler                            * *
// ************************************************************************
//@-at
//@@c

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
static struct work_struct lud_handle_q;
#else
static struct tq_struct lud_handle_q;
#endif

#define MAX_HANDLERS	20
#define LUD_Q_SIZE	32

static struct {
	LUD_DEV* dev;
	unsigned int major;
	LUD_COMMAND  com;
} lud_Q[LUD_Q_SIZE];
static int	lud_rpos = 0;
static int	lud_wpos = 0;

static void lud_q_handler(void* unused)
{
	DEB(4,printk(KERN_DEBUG "lud_q_handler r=%d w=%d\n",lud_rpos,lud_wpos));
	while (lud_rpos != lud_wpos) {
		LUD_DEV* dev = lud_Q[lud_rpos].dev;
		int major = lud_Q[lud_rpos].major;
		LUD_COMMAND* tc =  &lud_Q[lud_rpos].com;
		lud_rpos = (lud_rpos+1)%LUD_Q_SIZE;
		if (lud_devices[dev->minor].lud_nodes[major]->handler) {
			DEB(5,printk(KERN_DEBUG "calling handler for %d\n",major));
			lud_devices[dev->minor].lud_nodes[major]->handler(tc,dev);
		}
	}
}
//@-node:lud_q_handler
//@+node:lud_q_host_comm

static inline void lud_q_host_comm(int mod,LUD_COMMAND* tc,LUD_DEV* luddev)
{
	if (unlikely(tc == NULL))
	{
		printk(KERN_EMERG "MAJOR BUG: Luddite ISR got tc=NULL!\n");
		return;
	}

	DEB(6,printk(KERN_DEBUG "lud_q_host_comm %x %x %d %d %d %d\n",readl((IOTYPE)&tc->command),
		readl((IOTYPE)&tc->reply.pointer),
		readl((IOTYPE)&tc->arg0),
		readl((IOTYPE)&tc->arg1),
		readl((IOTYPE)&tc->arg2),
		readl((IOTYPE)&tc->arg3)));
	DEB(6,printk(KERN_DEBUG "from mod %d luddev %p: h=%lx,t=%lx,s=%lx\n",mod,luddev,luddev->host_comm_addr,luddev->target_comm_addr,luddev->scratch));

	/* Optimize the two most common messages, by handling them in the
	 * ISR.  This saves the cost of a context switch to ksoftirqd,
	 * that would otherwise be needed.
	 */
	if (unlikely(readl((IOTYPE)&tc->command) == LUD_SYS_REQUEST_ACK))
	{
		int need_free;
		LUD_DEV_LINK* tm = (LUD_DEV_LINK*)token_to_addr(readl((IOTYPE)&tc->reply.pointer));
		if (unlikely(tm == NULL))
		{
			printk(KERN_EMERG "BUG: Luddite ISR got LUD_REQUEST_ACK with reply=NULL!\n");
			return;
		}
		tm->retval0 = readl((IOTYPE)&tc->arg0);
		tm->retval1 = readl((IOTYPE)&tc->arg1);
		tm->retval2 = readl((IOTYPE)&tc->arg2);
		tm->retval3 = readl((IOTYPE)&tc->arg3);
		tm->status = LUD_REQUEST_DONE;
		need_free = atomic_dec_and_test(&tm->freecounter);
		wake_up_interruptible(&tm->q);
		if (likely(!need_free))
		{
			/* Finished handling (this is the normal case) */
			token_free(readl((IOTYPE)&tc->reply.pointer));
			return;
		}
		/*
		 * Need to free the link.
		 *
		 * Can't kfree() here since we're in interrupt context, and
		 * kfree uses a semaphore.
		 *
		 * So fall through, and redirect the message to the system
		 * handler.
		 */
		mod = LUD_SYSTEM_MAJOR;
		/*
		 * The message will be queued and sent to the system handler.
		 * The system handler will unconditionally free the link.
		 */
	}
	else if (unlikely( (readl((IOTYPE)&tc->command) == LUD_SYS_XFER_BUFFER)
	                && (mod == LUD_SYSTEM_MAJOR) ) )
	{
		/* Xfer buffer - common, so special-case for speed. */
		unsigned long irqstate;
		int need_free;
		LUD_CHANNEL * tmc = (LUD_CHANNEL *)token_to_addr(readl((IOTYPE)&tc->arg3));
		if (unlikely(tmc == NULL))
		{
			printk(KERN_EMERG "BUG: Luddite ISR got LUD_SYS_XFER_BUFFER with tmc=NULL!\n");
			return;
		}
		spin_lock_irqsave(&tmc->lock, irqstate);
		if (unlikely(tmc->pendingrequests==0)) {
			spin_unlock_irqrestore(&tmc->lock, irqstate);
			printk("lud-core: irq with pendingrequests==0\n");
			return;
		}
		tmc->addr = lud_remap_addr(luddev, readl((IOTYPE)&tc->arg0));
		tmc->used = readl((IOTYPE)&tc->arg1);
		tmc->max  = readl((IOTYPE)&tc->arg2);
		tmc->pendingrequests--;
		wake_up(&tmc->queue);
		need_free = ((tmc->refcount == 0) && (tmc->pendingrequests == 0));
		spin_unlock_irqrestore(&tmc->lock, irqstate);
		if (likely(!need_free))
		{
			/* Finished handling (this is the normal case) */
			token_free(readl((IOTYPE)&tc->arg3));
			return;
		}
		/*
		 * Need to free the channel.
		 *
		 * Can't kfree() here since we're in interrupt context, and
		 * kfree uses a semaphore.
		 *
		 * So fall through.
		 *
		 * The message will be queued and sent to the system handler.
		 * The system handler will unconditionally free the channel.
		 */
	}

	lud_Q[lud_wpos].dev = luddev;
	lud_Q[lud_wpos].major = mod;
	lud_Q[lud_wpos].com.command = readl((IOTYPE)&tc->command);
	lud_Q[lud_wpos].com.reply.cpu = readl((IOTYPE)&tc->reply.cpu);
	lud_Q[lud_wpos].com.reply.pointer = readl((IOTYPE)&tc->reply.pointer);
	lud_Q[lud_wpos].com.arg0 = readl((IOTYPE)&tc->arg0);
	lud_Q[lud_wpos].com.arg1 = readl((IOTYPE)&tc->arg1);
	lud_Q[lud_wpos].com.arg2 = readl((IOTYPE)&tc->arg2);
	lud_Q[lud_wpos].com.arg3 = readl((IOTYPE)&tc->arg3);

	lud_wpos = (lud_wpos+1)%LUD_Q_SIZE;

	if (unlikely(lud_wpos == lud_rpos)) {
		printk(KERN_EMERG "lud: command queue overflow!\n");
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
	schedule_work(&lud_handle_q);
#else
	schedule_task(&lud_handle_q);
#endif
}
//@-node:lud_q_host_comm
//@+node:lud_int_handler

void lud_int_handler(LUD_DEV* luddev)
{
	int mod;
	LUD_COMMAND* tc;

	switch (luddev->state) {
	case LUD_RUNNING:
		luddev->state = LUD_LUDDITE_STARTING;
		luddev->status = lud_remap_addr(luddev,readl((IOTYPE)luddev->host_comm_vec));
		luddev->scratch = luddev->status + LUD_CACHE_ROW_SIZE;
		break;
	case LUD_LUDDITE_STARTING:
	case LUD_LUDDITE:
		mod = LUD_NODE_FROM_HANDLE(readl((IOTYPE)luddev->host_comm_vec));
		tc = (LUD_COMMAND*)luddev->host_comm_addr;
		lud_q_host_comm(mod,tc,luddev);
		break;
	default:
		printk("lud %d ignoring interrupt\n",luddev->minor);
		return;		/* host_comm_vec not setup, so return */
		break;
	}
	writel(0,(IOTYPE)luddev->host_comm_vec);
}
//@-node:lud_int_handler
//@+node:show_dev_dev
/* ************************************************************************ *
 * Other functions...                                                       *
 * ************************************************************************ */

#ifdef CONFIG_SYSFS
static ssize_t show_dev_dev(struct class_device *dev, char *buf)
{
	char *str = buf;
	LUD_DEV* t = dev->class_data;
    str += sprintf(str,"%d:%d\n", MAJOR(t->cdev.dev),MINOR(t->cdev.dev));
    return (str - buf);
}
//@-node:show_dev_dev
//@+node:register_luddite_device
static CLASS_DEVICE_ATTR(dev,S_IRUGO,show_dev_dev,NULL);

static ssize_t show_dev_state(struct class_device *dev, char *buf)
{
    char *str = buf;
    LUD_DEV* t = dev->class_data;
	switch (t->state) {
		case LUD_UNDEFINED:
			str += sprintf(str,"undefined\n");
			break;
		case LUD_RESET:
			str += sprintf(str,"reset\n");
			break;
		case LUD_RUNNING:
			str += sprintf(str,"running\n");
			break;
		case LUD_LUDDITE_STARTING:
			str += sprintf(str,"luddite_starting\n");
			break;
		case LUD_LUDDITE:
			str += sprintf(str,"luddite\n");
			break;
		case LUD_CRASHED:
			str += sprintf(str,"crashed\n");
			break;
		default:
			str += sprintf(str,"unknown\n");
			break;
	}
        return (str - buf);
}

static CLASS_DEVICE_ATTR(state,S_IRUGO,show_dev_state,NULL);

static ssize_t show_dev_membase(struct class_device *dev, char *buf)
{
    char *str = buf;
	LUD_DEV* t = dev->class_data;
    str += sprintf(str,"%p\n", (void*)t->sdram_phy);
    return (str - buf);
}
static CLASS_DEVICE_ATTR(membase,S_IRUGO,show_dev_membase,NULL);

static ssize_t show_dev_memlen(struct class_device *dev, char *buf)
{
    char *str = buf;
	LUD_DEV* t = dev->class_data;
    str += sprintf(str,"0x%d\n",t->sdram_size);
    return (str - buf);
}
static CLASS_DEVICE_ATTR(memlen,S_IRUGO,show_dev_memlen,NULL);

static ssize_t show_dev_type(struct class_device *dev, char *buf)
{
    char *str = buf;
	LUD_DEV* t = dev->class_data;
    str += sprintf(str,"%s\n",t->name);
    return (str - buf);
}
static CLASS_DEVICE_ATTR(type,S_IRUGO,show_dev_type,NULL);

static ssize_t show_dev_frequency(struct class_device *dev, char *buf)
{
    char *str = buf;
	LUD_DEV* t = dev->class_data;
    str += sprintf(str,"%d\n",t->freq);
    return (str - buf);
}
static CLASS_DEVICE_ATTR(frequency,S_IRUGO,show_dev_frequency,NULL);

static ssize_t show_dev_dpl(struct class_device *dev, char *buf)
{
    char *str = buf;
    LUD_DEV* t = dev->class_data;
    str += sprintf(str,"%d\n",lud_devices[t->minor].dprintk_level);
    return (str - buf);
}

static ssize_t scan_dev_dpl(struct class_device *dev,const char *buf,size_t count)
{
	int dpl = -1;
	LUD_DEV* t = dev->class_data;

	sscanf(buf,"%d",&dpl);

	if (dpl>=0) {
		lud_devices[t->minor].dprintk_level = dpl;
	}

	return count;
}

static CLASS_DEVICE_ATTR(dpl,S_IRUGO|S_IWUSR,show_dev_dpl,scan_dev_dpl);
#endif

int register_luddite_device(LUD_DEV* dev)
{
#ifdef CONFIG_SYSFS
	struct cdev* chrdev;
#endif
	int i;
	int index = -1;
	char procname[128];

	if (unlikely(down_interruptible(&tms_sem))) {
		return -ERESTARTSYS;
	}

	for (i=0;i<MAX_LUD_DEV;i++) {
		if (lud_devices[i].luddev == NULL) {
			index = i;
			break;
		}
	}

	if (index<0) {
		up(&tms_sem);
		return -ENODEV;
	}

	lud_devices[index].dprintk_level = DEF_DPRINTK_LEVEL;
	lud_devices[index].luddev = dev;
	up(&tms_sem);
	sema_init(&lud_devices[index].sem,1);
	sema_init(&lud_devices[index].mod_sem,1);
	sema_init(&lud_devices[index].node_sem,1);
	init_waitqueue_head(&dev->start_queue);

	dev->major = LUD_MAJOR;
	dev->minor = index;

	initstr(index);

	create_proc_files(dev);

	sprintf(procname,"%d",dev->minor);

#ifdef CONFIG_SYSFS
	chrdev = &dev->cdev;
	cdev_init(chrdev,&lud_dev_fops);
	chrdev->owner = THIS_MODULE;
	kobject_set_name(&chrdev->kobj,"lud%d",index);
	if (cdev_add(chrdev, MKDEV(LUD_MAJOR,index), 1)) {
		printk("Error adding cdev %p\n",chrdev);
	}
	/* Add device */
	memset(&dev->ldev,0,sizeof(struct device));
	dev->ldev.bus = NULL;
	dev->ldev.parent = dev->parent;
	dev->ldev.release = lud_dev_release;
	dev->ldev.devt = MKDEV(LUD_MAJOR, index);
	snprintf(dev->ldev.bus_id, BUS_ID_SIZE, "lud%d",index);
	if (device_register(&dev->ldev))
	{
 		printk("Error registering device.\n");
	}
	dev->classdev = class_device_create(&lud_class, NULL, MKDEV(LUD_MAJOR,index), &dev->ldev, "lud%d", index);
	dev->classdev->class_data = dev;

	class_device_create_file(dev->classdev,&class_device_attr_dev);
	class_device_create_file(dev->classdev,&class_device_attr_dpl);
	class_device_create_file(dev->classdev,&class_device_attr_type);
	class_device_create_file(dev->classdev,&class_device_attr_state);
	class_device_create_file(dev->classdev,&class_device_attr_membase);
	class_device_create_file(dev->classdev,&class_device_attr_memlen);
	class_device_create_file(dev->classdev,&class_device_attr_frequency);
#else
	dev->devfs_directory = devfs_mk_dir(luddite_devfs_dir, procname, NULL);
	if (!dev->devfs_directory) {
		unregister_luddite_device(dev);
		return -ENOMEM;
	}

	dev->devfs_handle = devfs_register(dev->devfs_directory,"lud",
					   DEVFS_FL_AUTO_DEVNUM,
					   0, 0,
					   S_IFCHR | S_IRUGO | S_IWUGO,
					   &lud_dev_fops,lud_devices[dev->minor].luddev);
	if (!dev->devfs_handle) {
		unregister_luddite_device(dev);
		return -ENOMEM;
	}
#endif

	printk(KERN_DEBUG "lud_core: Registering devfs device %d/lud\n",dev->minor);

	if (dev->init) dev->init(dev);

	return 0;
}
//@-node:register_luddite_device
//@+node:unregister_luddite_device

void unregister_luddite_device(LUD_DEV* dev)
{
	int i;
	int index = -1;

	for (i=0;i<MAX_LUD_DEV;i++) {
		if (lud_devices[i].luddev == dev) {
			index = i;
			break;
		}
	}

	if (index<0) {
		return;
	}

	remove_proc_files(dev);

	/* Delete system links. */
	lud_devices[index].lud_nodes[LUD_SYSTEM_MAJOR]->handler = NULL;
	lud_devices[index].lud_nodes[LUD_LOADER_MAJOR]->handler = NULL;
	lud_devices[index].system = NULL;
	lud_devices[index].loader = NULL;

	/* Delete all nodes on this dead TriMedia */
	for (i = 0; i < LUD_MAX_NODES; i++) {
		del_node(index, i);
	}

#ifdef CONFIG_SYSFS
	class_device_remove_file(dev->classdev,&class_device_attr_type);
	class_device_remove_file(dev->classdev,&class_device_attr_state);
	class_device_remove_file(dev->classdev,&class_device_attr_membase);
	class_device_remove_file(dev->classdev,&class_device_attr_memlen);
	class_device_remove_file(dev->classdev,&class_device_attr_frequency);
	class_device_remove_file(dev->classdev,&class_device_attr_dpl);
	class_device_remove_file(dev->classdev,&class_device_attr_dev);
	device_unregister(&dev->ldev);
	cdev_del(&dev->cdev);
#else
	if (dev->devfs_handle)
		devfs_unregister(dev->devfs_handle);
	if (dev->devfs_directory)
		devfs_unregister(dev->devfs_directory);
#endif

	cleanstr(index);

	down(&tms_sem);
	lud_devices[index].luddev = NULL;
	up(&tms_sem);
}
//@-node:unregister_luddite_device
//@+node:init_module
module_param(lud_debug, uint, 0);
module_param(clear_target, uint, 0);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Philips Research");
MODULE_DESCRIPTION("Luddite core");


int init_module(void)
{
	printk(KERN_INFO "Luddite version %d, exported %s\n", LUDDITE_VERSION,
		EXPORTED_DATE);

	sema_init(&tms_sem,1);

	init_proc_system();

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
	SET_MODULE_OWNER(&lud_dev_fops);
	SET_MODULE_OWNER(&lud_node_fops);
#endif

#ifdef CONFIG_SYSFS
	class_register(&lud_class);

	if (register_chrdev_region(MKDEV(LUD_MAJOR,0), 256, "luddite") < 0) {
		printk(KERN_ERR "lud: Unable to register chrdev region.\n");
		return -EBUSY;
	}
#else
	if (devfs_register_chrdev(LUD_MAJOR, LUD_NAME,&lud_dev_fops) == -EBUSY) {
		printk(KERN_ERR "lud: Unable to register device.\n") ;
		return -EBUSY;
	}
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,21))
	INIT_WORK(&lud_handle_q, (void*)lud_q_handler);
#else
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
	INIT_WORK(&lud_handle_q, (void*)lud_q_handler, NULL);
#else
	lud_handle_q.routine = lud_q_handler;
	lud_handle_q.data = NULL;
#endif
#endif


#ifndef CONFIG_SYSFS
	luddite_devfs_dir = devfs_mk_dir (NULL, "luddite", NULL);
#endif

	return 0;
}
//@-node:init_module
//@+node:cleanup_module

void cleanup_module(void)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
	if (MOD_IN_USE) {
		return;
	}
#endif

#ifdef CONFIG_SYSFS
	class_unregister(&lud_class);
	unregister_chrdev_region(MKDEV(LUD_MAJOR,0),256);
#else
	devfs_unregister_chrdev(LUD_MAJOR, LUD_NAME);

	if (luddite_devfs_dir)
		devfs_unregister(luddite_devfs_dir);
#endif

	clean_proc_system();
}
//@-node:cleanup_module
//@+node:lud_node_match

static int lud_node_match(char* mod_name,char* name)
{
	int ln = strlen(name);
	if (name[ln-1] == '-') {
		if (mod_name && (strncmp(mod_name,name,ln)==0))
			return 1;
	}
	if (mod_name && (strcmp(mod_name ,name)==0))  {
			return 1;
	}
	return 0;
}
//@-node:lud_node_match
//@+node:lud_get_node_from_link
//@+at
//@nonl
// * * Get a pointer to the node structure given a LUD_DEV_LINK*
//@-at
//@@c

struct LUD_NODE_DEV* lud_get_node_from_link(LUD_DEV_LINK* l)
{
	return lud_devices[l->minor].lud_nodes[l->major];
}
//@-node:lud_get_node_from_link
//@+node:lud_get_node_link
//@+at
//@nonl
// * * Get a node with a specified name. * * If the name specified ends in
// '-', then a prefix search is performed. * This allows the pattern "mpg-" to
// match "mpg-0", "mpg-1", etc. * (Note that it also matches "mpg-extra-0", so
// be careful when choosing node * names.) * * @param last Ignored - FIXME. *
// @param name The node name or name prefix. * @param handler The handler for
// messages from TriMedia to Linux on that node. * * @todo FIXME 'last'
// paramater is unused - should be removed?
//@-at
//@@c

LUD_DEV_LINK*  lud_get_node_link(LUD_DEV_LINK* last,char* name,lud_handler_func_t handler)
{
	int i,j;
	LUD_DEV_LINK *result;

	for (j=0;j<MAX_LUD_DEV;j++) {
		if (lud_devices[j].luddev == NULL)
			continue;
		for (i=0;i<LUD_MAX_NODES;i++) {
			if (lud_devices[j].lud_nodes[i] && lud_node_match(lud_devices[j].lud_nodes[i]->name,name)) {
				if (lud_devices[j].lud_nodes[i]->handler)
					continue;
				result = lud_new_dev(i,j,handler);
				if (likely(result != NULL)) {
					INC_LUD_MOD_USE_COUNT(result);
				}
				return result;
			}
		}
	}
	return NULL;
}
//@-node:lud_get_node_link
//@+node:lud_get_node_link_from_handle
//@+at
//@nonl
// * * Get a node, if you have recieved the node's handle from a TriMedia-side
// process. * * @param parent The LUD_DEV_LINK you used to get the node
// handle. This allows *        lud-core to determine which TriMedia CPU is
// being used. *        Must not be NULL. * @param handle The node handle for
// the node you want, passed from the TriMedia *        side via some
// module-specific mechanism. * @param handler The handler function * @return
// The node link, or NULL on error (invalid handle).
//@-at
//@@c

LUD_DEV_LINK* lud_get_node_link_from_handle(LUD_DEV_LINK* parent, int handle, lud_handler_func_t handler)
{
	LUD_DEV_LINK *result;

	if (unlikely(parent == NULL))
	{
		/* Should never get here - caller must ensure 'parent' is valid. */
		printk(KERN_CRIT "Bug detected at %s:%d\n", __FILE__, __LINE__);
		BUG();
		return NULL;
	}

	if (unlikely(parent->minor < 0)
	 || unlikely(parent->minor >= MAX_LUD_DEV))
	{
		/* Should never get here - 'parent' is corrupted. */
		printk(KERN_CRIT "Bug detected at %s:%d - %d not in range 0 to %d\n", __FILE__, __LINE__, parent->major, MAX_LUD_DEV);
		return NULL;
	}

	if (unlikely(lud_devices[parent->minor].luddev == NULL))
	{
		/* Should never get here.
		 *
		 * Looks like the relevent TriMedia module was removed,
		 * but we should have it marked 'in use' via the 'parent' link.
		 */
		printk(KERN_CRIT "Bug detected at %s:%d\n", __FILE__, __LINE__);
		BUG();
		return NULL;
	}

	if (unlikely(handle <= 0) || unlikely(handle >= LUD_MAX_NODES))
	{
		/* Invalid handle */
		printk(KERN_CRIT "Bug detected at %s:%d\n", __FILE__, __LINE__);
		return NULL;
	}

	if (unlikely(!lud_devices[parent->minor].lud_nodes[handle]))
	{
		/* No such node */
		printk(KERN_CRIT "Bug detected at %s:%d\n", __FILE__, __LINE__);
		return NULL;
	}

	if (unlikely(lud_devices[parent->minor].lud_nodes[handle]->handler != NULL))
	{
		/* Node is already allocated */
		printk(KERN_CRIT "Bug detected at %s:%d\n", __FILE__, __LINE__);
		return NULL;
	}

	result = lud_new_dev(handle,parent->minor,handler);
	if (likely(result != NULL)) {
		INC_LUD_MOD_USE_COUNT(result);
	}
	return result;
}
//@-node:lud_get_node_link_from_handle
//@+node:lud_free_node_link

void lud_free_node_link(LUD_DEV_LINK* tm)
{
	if ((tm->minor < 0) || (tm->minor >= MAX_LUD_DEV) ||
	    (tm->major < 1) || (tm->major >= LUD_MAX_NODES)) {
		printk(KERN_CRIT "Luddite bug: lud_free_node_link(invalid link).\n");
		return;
	}

	if (lud_devices[tm->minor].lud_nodes[tm->major]->handler == NULL) {
		printk(KERN_CRIT "Luddite bug: lud_free_node_link(freed link).\n");
		return;
	}

	lud_devices[tm->minor].lud_nodes[tm->major]->handler = NULL;
	DEC_LUD_MOD_USE_COUNT(tm);
}
//@-node:lud_free_node_link
//@-others


/**
 * @}
 */


EXPORT_SYMBOL(register_luddite_device);
EXPORT_SYMBOL(lud_int_handler);
EXPORT_SYMBOL(unregister_luddite_device);
EXPORT_SYMBOL(lud_devices);
EXPORT_SYMBOL(lud_get_node_link);
EXPORT_SYMBOL(lud_get_node_link_from_handle);
EXPORT_SYMBOL(lud_get_node_from_link);
EXPORT_SYMBOL(lud_free_node_link);
EXPORT_SYMBOL(lud_free_link);
EXPORT_SYMBOL(lud_clone_link);
EXPORT_SYMBOL(lud_send_command);
EXPORT_SYMBOL(lud_send_request);
EXPORT_SYMBOL(lud_send_ack);
EXPORT_SYMBOL(lud_remap_addr);
EXPORT_SYMBOL(lud_wakeup);

//@-node:@file host/base.c
//@-leo
