//@+leo-ver=4
//@+node:@file host/channels.c
//@@language c
//@<< lud-base #includes >>
//@+node:<< lud-base #includes >>
#ifdef CONFIG_SMP
#define __SMP__
#endif

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/page.h>
#include <asm/semaphore.h>
#include <asm/atomic.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>

#include <linux/module.h>

#include <linux/version.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/kmod.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#include <linux/device.h>
#include <linux/cdev.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18))
#include <linux/compiler.h>
#else
#define likely(x) (x)
#define unlikely(x) (x)
#endif /* LINUX_VERSION_CODE */

#include <stdarg.h>


#include "compat.h"
#include "lud_mmio.h"
#include "luddite-host.h"
#include "luddite/version.h"
#include "luddite/os.h"
#include "exported.h"

#include "core.h"
#include "procfs.h"
#include "node.h"
#include "helpers.h"
#include "channels.h"
//@-node:<< lud-base #includes >>
//@nl
//@+others
//@+node:lud_new_channel

/* ************************************************************************ *
 * Xfer functions                                                           *
 * ************************************************************************ */

/**
 * Create a new channel.  (A "channel" on the host side, and an "xfer"
 * on the target, are the same thing).
 *
 * Will send a LUD_SYS_OPEN_CHANNEL request to the target, with the
 * channel ID as the first argument.
 *
 * @param tm The link to use.  Will be cloned by this function.
 * @param direction The channel direction.
 * @return A new LUD_CHANNEL, or NULL.  Free with lud_free_channel().
 */
LUD_CHANNEL* lud_new_channel(LUD_DEV_LINK* tm,int direction)
{
	LUD_CHANNEL* tmc = kmalloc(sizeof(LUD_CHANNEL),GFP_KERNEL);
	if (!tmc) {
		return NULL ;
	}
	tmc->luddev = lud_clone_link(tm);
	if (unlikely(tmc->luddev == NULL)) {
		kfree(tmc);
		return NULL;
	}

	lud_send_command(tmc->luddev,LUD_SYS_OPEN_CHANNEL,direction,0,0,0);

	tmc->arg = direction;
	tmc->addr = 0;
	tmc->used = 0;
	tmc->max  = 0;

	spin_lock_init(&tmc->lock);
	tmc->pendingrequests = 0;
	tmc->refcount = 1;
	init_waitqueue_head(&tmc->queue);

	return tmc;
}
//@-node:lud_new_channel
//@+node:lud_new_write_channel

LUD_CHANNEL* lud_new_write_channel(LUD_DEV_LINK* tm)
{
	return lud_new_channel(tm,LUD_SYS_WRITE_CHANNEL);
}
//@-node:lud_new_write_channel
//@+node:lud_new_read_channel
LUD_CHANNEL* lud_new_read_channel(LUD_DEV_LINK* tm)
{
	return lud_new_channel(tm,LUD_SYS_READ_CHANNEL);
}
//@-node:lud_new_read_channel
//@+node:lud_free_channel
//@+at
//@nonl
// * * Closes and frees a channel. * * @param tmc The channel. May be NULL.
//@-at
//@@c

void lud_free_channel(LUD_CHANNEL* tmc)
{
	unsigned long irqstate;
	int need_free;

	if (unlikely(tmc == NULL))
	{
		return;
	}

	/* check to make sure the Target side is still OK ...
	 * it might have been reset using the lud-clean command */
	if (likely(tmc->luddev->dev->state == LUD_LUDDITE)) {
		/* Tell the TriMedia side that the Xfer is being closed.
		 * If the Xfer is waiting for data from the TriMedia (i.e.
		 * if tmc->pendingrequests != 0) then this tells the TriMedia
		 * to send a NULL buffer immediately, not to wait for valid
		 * data.  If the TriMedia has already responded to all requests,
		 * then it will ignore this command.
		 */
		lud_send_request(tmc->luddev,LUD_SYS_CLOSE_CHANNEL,
						tmc->arg,0,0,0);
	}

	spin_lock_irqsave(&tmc->lock, irqstate);
	if (unlikely(tmc->refcount == 0))
	{
		spin_unlock_irqrestore(&tmc->lock, irqstate);
		printk(KERN_EMERG "Luddite double-free bug - lud_free_channel(%p) at "
		       HERE_STRING_FORMAT "\n", (void *)tmc, HERE_STRING_ARGS);
		return;
	}
	tmc->refcount--;
	need_free = ((tmc->refcount == 0) && (tmc->pendingrequests == 0));
	spin_unlock_irqrestore(&tmc->lock, irqstate);

	if (likely(need_free))
	{
		lud_free_link(tmc->luddev);
		kfree(tmc);
	}
}
//@-node:lud_free_channel
//@+node:lud_read_channel
//@+at
//@nonl
// * * Read from a channel. * * This is a low-level version. It is intended
// for use in implementing * simple Linux character devices, where file
// operations map directly * to an Xfer, and where you want to provide the
// guarantees required * by select(). If you are writing a more complex kernel
// module, or * if you don't want to support select() or poll(), then you will
// * probably find it easier use lud_read_channel_buffered() instead. * * This
// function will send at most one request to the TriMedia. * * If there is
// already data buffered, then it will merely return * that data, and it will
// do so immediately. * * If there is no data buffered, it will send the
// TriMedia a * single request for data, and return as much as possible of *
// that data, up to the limit specified. Extra data will be * buffered. * *
// This function may be interrupted by a signal (unless nonblock is set). * *
// Thread safety: Caller must ensure that they do not attempt multiple *
// simultaneous reads from the same channel. You may simultaneously read *
// from and poll a channel. You must ensure that all operations on the *
// channel have returned before you free the channel. * * @param tmc   The
// channel to read from. * @param b    The buffer to store read data in. *
// @param n    The size of the buffer, in bytes. * @param kernel  Nonzero if
// the buffer is in kernel space, zero if *         the buffer is in user
// space. * @param nonblock If nonzero, the TriMedia will be told not to
// block. *         It will fail rather than waiting for the TriMedia
// *         code to produce more data. Note that this does not *         stop
// this function from blocking, it must still wait *         for a reply from
// the TriMedia. * @return     The number of bytes read. Likely to be less
// than *         the requested size. Will not be zero unless nonblock
// *         is set. Error codes: *         - -ERESTARTSYS if interrupted by a
// signal. *         - -EBADF if tmc is NULL. *         - 0 if n<=0.
//@-at
//@@c

int lud_read_channel(LUD_CHANNEL* tmc,unsigned char* b,int n,int kernel,int nonblock)
{
	int count;
	unsigned long irqstate;
	unsigned long addr;
	int need_more_data;

	if (unlikely(tmc == NULL))
		return -EBADF;

	if (unlikely(n <= 0))
		return 0;

	//printk("Linux-side lud_read_channel() entered - want %d bytes\n", n);

	spin_lock_irqsave(&tmc->lock, irqstate);

	/* First, if we don't already have some data, get some from the TriMedia. */
	if (tmc->used == 0) {
		if (likely(tmc->pendingrequests == 0)) {
			tmc->pendingrequests++;
			spin_unlock_irqrestore(&tmc->lock, irqstate);
			lud_send_command(tmc->luddev,LUD_SYS_READ_CHANNEL,n,0,tmc->arg,addr_to_token(tmc));
		}
		else
		{
			spin_unlock_irqrestore(&tmc->lock, irqstate);
		}
		if (unlikely(nonblock)) {
			/* Don't want to wait for data */
			return 0;
		}
		/* Wait for a reply - i.e. for ((tmc->used != 0) ||
		 * (tmc->pendingrequests == 0)) to become true.
		 * This is slightly complicated by the need to:
		 * 1) Lock and unlock the spinlock, and
		 * 2) Don't unlock the spinlock if the event is
		 *    true.
		 */
		if (unlikely(wait_event_interruptible(tmc->queue,
				({
					int retval;
					spin_lock_irqsave(&tmc->lock, irqstate);
					retval = ((tmc->used != 0) || (tmc->pendingrequests == 0));
					if (unlikely(!retval))
					{
						spin_unlock_irqrestore(&tmc->lock, irqstate);
					}
					retval;
				})
			)))
		{
			/* Spinlock is *NOT LOCKED* here */
			/* We were interrupted (by Ctrl-C or another signal) */
			return -ERESTARTSYS;
		}
		/* Spinlock is *LOCKED* here */
		if (unlikely(tmc->used == 0))
		{
			/* Should not get here. */
			spin_unlock_irqrestore(&tmc->lock, irqstate);
			printk(KERN_EMERG "Shouldn't have got to " HERE_STRING_FORMAT "\n", HERE_STRING_ARGS);
			return 0;
		}
	}

	/* Spinlock is *LOCKED* here */

	/* Now we have some data, pass it on to the caller. */
	count = (n > tmc->used) ? tmc->used : n;
	//printk("Linux-side lud_read_channel() copying %d bytes\n", count);
	addr = tmc->addr;
	tmc->used -= count;
	tmc->addr += count;

	need_more_data = ((tmc->used == 0) && (tmc->pendingrequests == 0));
	if (need_more_data) {
		tmc->pendingrequests++;
	}

	spin_unlock_irqrestore(&tmc->lock, irqstate);

	if (kernel) {
		memcpy(b,(void*)addr,count);
	} else {
		if (unlikely(copy_to_user_fromio(b,(void*)addr,count))) {
			count = -EFAULT;
		}
	}


	if (need_more_data) {
		lud_send_command(tmc->luddev,LUD_SYS_READ_CHANNEL,1,0,tmc->arg,addr_to_token(tmc));
	}

	//printk("Linux-side lud_read_channel() OK - %d bytes\n", count);
	return count;
}
//@-node:lud_read_channel
//@+node:lud_read_channel_buffered
//@+at
//@nonl
// * * Read from a channel. This is a higher-level function, if you * want a
// low-level one then use lud_read_channel(). * * This function will wait
// until the requested amout of data is * available. * * This function may be
// interrupted by a signal (unless nonblock is set). * * Thread safety: Caller
// must ensure that they do not attempt multiple * simultaneous reads from the
// same channel. You may simultaneously read * from and poll a channel. You
// must ensure that all operations on the * channel have returned before you
// free the channel. * * @param tmc   The channel to read from. * @param b
// The buffer to store read data in. * @param n    The size of the buffer, in
// bytes. * @param kernel  Nonzero if the buffer is in kernel space, zero if
// *         the buffer is in user space. * @param nonblock If nonzero, the
// TriMedia will be told not to block. *         It will fail rather than
// waiting for the TriMedia *         code to produce more data. Note that
// this does not *         stop this function from blocking, it must still
// wait *         for a reply from the TriMedia. * @return     The number of
// bytes read. Will not be less than *         the requested size unless
// nonblock is set or an *         error occurs. Error codes: *         -
// -ERESTARTSYS if interrupted by a signal before any *          data can be
// read. (If some data is read, then that *          will be returned even if
// a later error occurs). *         - -EBADF if tmc is NULL. *         - 0 if
// n<=0.
//@-at
//@@c

int lud_read_channel_buffered(LUD_CHANNEL* tmc,unsigned char* b,int n,int kernel,int nonblock)
{
	int todo = n;

	//printk("Linux-side lud_read_channel_buffered() entered - want %d bytes\n", n);

	if (unlikely(n <= 0)) {
		return 0;
	}

	while (todo > 0) {
		int count = lud_read_channel(tmc,b,todo,kernel,nonblock);
		if (unlikely(count <= 0)) {
			/* Error (or count==0, which means nonblock mode and no data
			 * available).  Return number of bytes read, if any, else
			 * error code.  This gives caller at least some chance of
			 * recovery.
			 */
			return ((todo != n) ? (n - todo) : count);
		}
		b += count;
		todo -= count;
	}

	//printk("Linux-side lud_read_channel_buffered() OK - %d bytes\n", n);

	return n;
}
//@-node:lud_read_channel_buffered
//@+node:lud_write_channel_flush
//@+at
//@nonl
// * * Send any pending data on the channel. This is used in conjunction *
// with lud_write_channel_buffered(). * * This function alway returns
// immediately. * This function may not be interrupted by signals. * * Thread
// safety: Caller must ensure that they do not attempt multiple * simultaneous
// writes (or flushes) to the same channel. You may * simultaneously write to
// and poll a channel. You must ensure that all * operations on the channel
// have returned before you free the channel. * * @param tmc   The channel to
// write to. * @return     Zero on success. (Does not indicate whether or not
// *         any data was actually present to be flushed). *         -EBADF if
// tmc is NULL
//@-at
//@@c

int lud_write_channel_flush(LUD_CHANNEL* tmc)
{
	unsigned long irqstate;

	if (unlikely(tmc == NULL))
		return -EBADF;

	if (unlikely(tmc->luddev->dev->state != LUD_LUDDITE)) {
		return 0;
	}

	spin_lock_irqsave(&tmc->lock, irqstate);

	if (likely(tmc->used != 0)) {
		int used = tmc->used;
		tmc->used = 0;
		tmc->addr = 0;
		tmc->pendingrequests++;
		spin_unlock_irqrestore(&tmc->lock, irqstate);
		lud_send_command(tmc->luddev,LUD_SYS_WRITE_CHANNEL,used,0,tmc->arg,addr_to_token(tmc));
	}
	else
	{
		spin_unlock_irqrestore(&tmc->lock, irqstate);
	}

	return 0;
}
//@-node:lud_write_channel_flush
//@+node:lud_write_channel_ex
//@+at
//@nonl
// * * Write data to the TriMedia across a channel. * * This is the
// implementation of lud_write_channel() and (most of) *
// lud_write_channel_buffered(), the desired behaviour can be * selected with
// a flag. * * This function may be interrupted by a signal (unless nonblock
// is set). * * Thread safety: Caller must ensure that they do not attempt
// multiple * simultaneous writes (or flushes) to the same channel. You may *
// simultaneously write to and poll a channel. You must ensure that all *
// operations on the channel have returned before you free the channel. * *
// @param tmc   The channel to write to. * @param b    The buffer of data to
// write. Must be non-NULL. * @param n    The amount of data to write, in
// bytes. * @param kernel  Nonzero if the buffer is in kernel space, zero if
// *         the buffer is in user space. Caller is responsible *         for
// validating user space buffers. * @param nonblock If nonzero, the TriMedia
// will be told not to block. *         It will fail rather than waiting for
// the TriMedia *         code to provide an empty buffer. Note that this
// *         does not stop this function from blocking, it must *
// still wait for a reply from the TriMedia. * @param flush  If nonzero, the
// data is always flushed after the write. * @return     The number of bytes
// sent. Likely to be less than *         the requested size. Will not be zero
// unless nonblock *         is set. Error codes: *         - -ERESTARTSYS if
// interrupted by a signal *         - -EBADF if tmc is NULL. *         - 0 if
// n<=0.
//@-at
//@@c

int lud_write_channel_ex(LUD_CHANNEL* tmc,const unsigned char* b,int n,int kernel,int nonblock,int flush)
{
	int copy;
	unsigned long irqstate;
	long dest;
	int need_flush;
	int used;

	if (unlikely(tmc == NULL))
		return -EBADF;

	if (unlikely(n <= 0))
		return 0;

	//printk("lud_write_channel_internal: nb=%d, n=%d\n", nonblock, n);

	spin_lock_irqsave(&tmc->lock, irqstate);

	/* Get a buffer if we haven't got one. */
	if (tmc->addr == 0) {
		int used = tmc->used;
		tmc->used = 0;
		if (tmc->pendingrequests == 0) {
			tmc->pendingrequests++;
			spin_unlock_irqrestore(&tmc->lock, irqstate);
			lud_send_command(tmc->luddev,LUD_SYS_WRITE_CHANNEL,used,0,tmc->arg,addr_to_token(tmc));
		}
		else
		{
			spin_unlock_irqrestore(&tmc->lock, irqstate);
		}
		if (nonblock) {
			/* Don't want to wait for a buffer */
			//printk("lud_write_channel_internal: nb=%d, n=%d, a=0  (nonblock fail)\n", nonblock, n);
			return 0;
		}
		/* Wait for a reply - i.e. for ((tmc->addr != 0) ||
		 * (tmc->pendingrequests == 0)) to become true.
		 * This is slightly complicated by the need to:
		 * 1) Lock and unlock the spinlock, and
		 * 2) Don't unlock the spinlock if the event is
		 *    true.
		 */
		if (unlikely(wait_event_interruptible(tmc->queue,
				({
					int retval;
					spin_lock_irqsave(&tmc->lock, irqstate);
					retval = ((tmc->addr != 0) || (tmc->pendingrequests == 0));
					if (unlikely(!retval))
					{
						spin_unlock_irqrestore(&tmc->lock, irqstate);
					}
					retval;
				})
			)))
		{
			/* Spinlock is *NOT LOCKED* here */
			/* We were interrupted (by Ctrl-C or another signal) */
			//printk("lud_write_channel_internal: nb=%d, n=%d, a=-1  (interrupted)\n", nonblock, n);
			return -ERESTARTSYS;
		}
		/* Spinlock is *LOCKED* here */
		if (unlikely(tmc->addr == 0))
		{
			/* Should not get here. */
			spin_unlock_irqrestore(&tmc->lock, irqstate);
			printk(KERN_EMERG "Shouldn't have got to " HERE_STRING_FORMAT "\n", HERE_STRING_ARGS);
			return 0;
		}
	}

	/* Spinlock is *LOCKED* here */

	copy = tmc->max - tmc->used; /* max # bytes to copy */
	if (likely(n < copy)) {
		copy = n;
	}
	/* copy = actual # bytes to copy */

	//printk("Linux-side lud_write_channel*() copying %d bytes\n", copy);

	dest = tmc->addr + tmc->used;
	tmc->used += copy;

	used = tmc->used;
	need_flush = ((used == tmc->max) || flush);
	if (likely(need_flush)) {
		tmc->used = 0;
		tmc->addr = 0;
		tmc->pendingrequests++;
	}

	spin_unlock_irqrestore(&tmc->lock, irqstate);

	if (kernel) {
		memcpy((void*)dest,b,copy);
	} else {
		if (unlikely(copy_from_user_toio((void*)dest,(void*)b,copy)))
			copy = -EFAULT;
	}

	if (likely(need_flush)) {
		/* Flush if we have a full buffer.
		 * In unbuffered mode, always flush. */
		//printk("Linux-side lud_write_channel*() is flushing %d/%d bytes\n", tmc->used, tmc->max);
		lud_send_command(tmc->luddev,LUD_SYS_WRITE_CHANNEL,used,0,tmc->arg,addr_to_token(tmc));
	}

	//printk("lud_write_channel_internal: nb=%d, n=%d, a=%d\n", nonblock, n, copy);

	return copy;
}
//@-node:lud_write_channel_ex
//@+node:lud_write_channel
//@+at
//@nonl
// * * Write data to the TriMedia across a channel. * * This is a low-level
// version. It is intended for use in implementing * simple Linux character
// devices, where file operations map directly * to an Xfer, and where you
// want to provide the guarantees required * by select(). If you are writing a
// more complex kernel module, or * if you don't want to support select() or
// poll(), then you will * probably find it easier use
// lud_write_channel_buffered() instead. * * The data is always flushed after
// the write. * * This function may be interrupted by a signal (unless
// nonblock is set). * * Thread safety: Caller must ensure that they do not
// attempt multiple * simultaneous writes (or flushes) to the same channel.
// You may * simultaneously write to and poll a channel. You must ensure that
// all * operations on the channel have returned before you free the channel.
// * * @param tmc   The channel to write to. * @param b    The buffer of data
// to write. Must be non-NULL. * @param n    The amount of data to write, in
// bytes. * @param kernel  Nonzero if the buffer is in kernel space, zero if
// *         the buffer is in user space. Caller is responsible *         for
// validating user space buffers. * @param nonblock If nonzero, the TriMedia
// will be told not to block. *         It will fail rather than waiting for
// the TriMedia *         code to provide an empty buffer. * @return     The
// number of bytes sent. Likely to be less than *         the requested size.
// Will not be zero unless nonblock *         is set. Error codes: *         -
// -ERESTARTSYS if interrupted by a signal. *         - -EBADF if tmc is NULL.
// *         - 0 if n<=0.
//@-at
//@@c

int lud_write_channel(LUD_CHANNEL* tmc,const unsigned char* b,int n,int kernel,int nonblock)
{
	return lud_write_channel_ex(tmc,b,n,kernel,nonblock,1);
}
//@-node:lud_write_channel
//@+node:lud_write_channel_buffered_ex
//@+at
//@nonl
// * * Write data to the TriMedia across a channel. * * This is a high-level
// version. * * Note that for efficiency, only full packets are sent - the
// channel * is not usually flushed after the write. * * This function may be
// interrupted by a signal (unless nonblock is set). * * Thread safety: Caller
// must ensure that they do not attempt multiple * simultaneous writes (or
// flushes) to the same channel. You may * simultaneously write to and poll a
// channel. You must ensure that all * operations on the channel have returned
// before you free the channel. * * @param tmc   The channel to write to. *
// @param b    The buffer of data to write. Must be non-NULL. * @param n
// The amount of data to write, in bytes. * @param kernel  Nonzero if the
// buffer is in kernel space, zero if *         the buffer is in user space.
// Caller is responsible *         for validating user space buffers. * @param
// nonblock If nonzero, the TriMedia will be told not to block. *         It
// will fail rather than waiting for the TriMedia *         code to provide an
// empty buffer. Note that this *         does not stop this function from
// blocking, it must *         still wait for a reply from the TriMedia.
// *         If this is set, there is also a limit of a single *
// packet sent to the TriMedia - if the data does *         not fit into a
// single packet, then only some *         of the data will be sent. * @param
// flush  If nonzero, the data is always flushed after the write. *
// @return     The number of bytes sent. Will not be less than *         the
// requested size unless nonblock is set or an *         error or signal
// occurs. *         Error codes: *         - -ERESTARTSYS if interrupted by a
// signal before *          any data could be written. (If some data is
// *          written, then the number of bytes written will be *
// returned even if a later error occurs). *         - -EBADF if tmc is NULL.
// *         - 0 if n<=0.
//@-at
//@@c

int lud_write_channel_buffered_ex(LUD_CHANNEL* tmc,const unsigned char* b,int n,int kernel,int nonblock,int flush)
{
	int todo = n;

	//printk("lud_write_channel_buffered(): %d bytes\n", n);

	if (unlikely(n <= 0)) {
		return 0;
	}

	while (todo > 0) {
		int count = lud_write_channel_ex(tmc,b,todo,kernel,nonblock,flush);
		if (unlikely(count < 0)) {
			/* Error (or count==0, which means nonblock mode and no buffer
			 * available).  Return number of bytes written, if any, else
			 * error code.  This gives caller at least some chance of
			 * recovery.
			 */
			//printk("lud_write_channel_buffered(): %d bytes, error %d\n", n, count);
			return ((todo != n) ? (n - todo) : count);
		}
		b += count;
		todo -= count;
	}

	//printk("lud_write_channel_buffered(): %d bytes done\n", n);
	return n;
}
//@-node:lud_write_channel_buffered_ex
//@+node:lud_write_channel_buffered
//@+at
//@nonl
// * * Write data to the TriMedia across a channel. * * This is a high-level
// version. * * Note that for efficiency, only full packets are sent - the
// channel * is not usually flushed after the write. * * This function may be
// interrupted by a signal (unless nonblock is set). * * Thread safety: Caller
// must ensure that they do not attempt multiple * simultaneous writes (or
// flushes) to the same channel. You may * simultaneously write to and poll a
// channel. You must ensure that all * operations on the channel have returned
// before you free the channel. * * @param tmc   The channel to write to. *
// @param b    The buffer of data to write. Must be non-NULL. * @param n
// The amount of data to write, in bytes. * @param kernel  Nonzero if the
// buffer is in kernel space, zero if *         the buffer is in user space.
// Caller is responsible *         for validating user space buffers. * @param
// nonblock If nonzero, the TriMedia will be told not to block. *         It
// will fail rather than waiting for the TriMedia *         code to provide an
// empty buffer. Note that this *         does not stop this function from
// blocking, it must *         still wait for a reply from the TriMedia.
// *         If this is set, there is also a limit of a single *
// packet sent to the TriMedia - if the data does *         not fit into a
// single packet, then only some *         of the data will be sent. *
// @return     The number of bytes sent. Will not be less than *         the
// requested size unless nonblock is set or an *         error or signal
// occurs. *         Error codes: *         - -ERESTARTSYS if interrupted by a
// signal before *          any data could be written. (If some data is
// *          written, then the number of bytes written will be *
// returned even if a later error occurs). *         - -EBADF if tmc is NULL.
// *         - 0 if n<=0.
//@-at
//@@c

int lud_write_channel_buffered(LUD_CHANNEL* tmc,const unsigned char* b,int n,int kernel,int nonblock)
{
	return lud_write_channel_buffered_ex(tmc,b,n,kernel,nonblock,0);
}
//@-node:lud_write_channel_buffered
//@+node:lud_poll_read_channel
//@+at
//@nonl
// * * Poll a channel. This is intended to help module authors * implement the
// poll() function on a linux device driver. * Apart from requiring an
// additional LUD_CHANNEL * argument, it * has the same syntax as the device
// driver poll() method. * * If there is already data buffered, then it will
// merely return * (POLLIN | POLLRDNORM). * * If there is no data buffered, it
// will send the TriMedia a * request for data, register the response wait
// queue with * poll_wait(filp, ..., wait), and return 0. * * This function
// always returns immediately - it does not block. * This function may not be
// interrupted by signals. * * Thread safety: Caller must ensure that they do
// not attempt multiple * simultaneous reads from the same channel. You may
// simultaneously read * from and poll a channel. You must ensure that all
// operations on the * channel have returned before you free the channel. * *
// @param tmc   The channel to read from. * @param filp   The file structure
// for poll_wait() * @param wait   The poll_table for poll_wait() *
// @return     (POLLIN | POLLRDNORM) if data is available. *         0 if no
// data is available or if tmc is NULL.
//@-at
//@@c

unsigned int lud_poll_read_channel(LUD_CHANNEL* tmc, struct file * filp, poll_table * wait)
{
	unsigned long irqstate;
	int retval = POLLIN | POLLRDNORM;
	int send_request = 0;

	if (unlikely(tmc == NULL))
		return 0;

	//printk("Linux-side lud_poll_read_channel() entered\n");

	poll_wait(filp, &tmc->queue, wait);

	spin_lock_irqsave(&tmc->lock, irqstate);

	if (tmc->used == 0) {
		/* No data available to read.  Ask the TriMedia to send some. */

		retval = 0;
		if (tmc->pendingrequests == 0) {
			tmc->pendingrequests++;
			send_request = 1;
		}
	}

	spin_unlock_irqrestore(&tmc->lock, irqstate);

	if (send_request) {
		lud_send_command(tmc->luddev,LUD_SYS_READ_CHANNEL,1,0,tmc->arg,addr_to_token(tmc));
	}

	return retval;
}
//@-node:lud_poll_read_channel
//@+node:lud_poll_write_channel
//@+at
//@nonl
// * * Poll a channel. This is intended to help module authors * implement the
// poll() function on a linux device driver. * Apart from requiring an
// additional LUD_CHANNEL * argument, it * has the same syntax as the device
// driver poll() method. * * If there is already a buffer available for
// writing, then * it will merely return (POLLOUT | POLLWRNORM). * * If there
// is no data buffered, it will send the TriMedia a * request for data,
// register the response wait queue with * poll_wait(filp, ..., wait), and
// return 0. * * This function always returns immediately - it does not block.
// * This function may not be interrupted by signals. * * Thread safety:
// Caller must ensure that they do not attempt multiple * simultaneous writes
// (or flushes) to the same channel. You may * simultaneously write to and
// poll a channel. You must ensure that all * operations on the channel have
// returned before you free the channel. * * @param tmc   The channel to read
// from. * @param filp   The file structure for poll_wait() * @param wait
// The poll_table for poll_wait() * @return     (POLLOUT | POLLWRNORM) if data
// is available. *         0 if no data is available or if tmc is NULL.
//@-at
//@@c

unsigned int lud_poll_write_channel(LUD_CHANNEL* tmc, struct file * filp, poll_table * wait)
{
	unsigned long irqstate;
	int retval = POLLOUT | POLLWRNORM;
	int send_request = 0;
	int used = 0;

	if (unlikely(tmc == NULL))
		return 0;

	//printk("Linux-side lud_poll_write_channel() entered\n");

	poll_wait(filp, &tmc->queue, wait);

	spin_lock_irqsave(&tmc->lock, irqstate);

	if (tmc->addr == 0) {
		/* No buffer available for write.  Ask the TriMedia to send one. */
		retval = 0;

		if (tmc->pendingrequests == 0) {
			used = tmc->used;
			tmc->used = 0;
			tmc->pendingrequests++;
			send_request = 1;
		}
	}

	spin_unlock_irqrestore(&tmc->lock, irqstate);

	if (send_request) {
		lud_send_command(tmc->luddev,LUD_SYS_WRITE_CHANNEL,used,0,tmc->arg,addr_to_token(tmc));
	}

	return retval;
}
//@-node:lud_poll_write_channel
//@+node:lud_wait_read_channel
//@+at
//@nonl
// * * Check if any data is available on this channel. If there isn't any, *
// it requests data, and optionally adds the current task to the wait * queue
// for the channel. * * This function is intended to help module authors
// implement polled * access to a number of channels, where integration with
// the poll() * system function is not appropriate. * * If there is data
// buffered, the amount of data presently available * for reading is
// returned.  * If there is no data buffered, it will send the TriMedia a *
// request for data. If a waitqueue_entry is provided, it is registered * with
// the response wait queue using. It will return 0. * * This function always
// returns immediately - it does not block. * This function may not be
// interrupted by signals. * * Thread safety: Caller must ensure that they do
// not attempt multiple * simultaneous reads from the same channel. You may
// simultaneously read * from and poll a channel. You must ensure that all
// operations on the * channel have returned before you free the channel. * *
// @param tmc   The channel to read from. * @param waitqueue_entry   The entry
// to add to this channel's wait queue * @return     +ve amount of data
// available *         0 if no data is available and
//@-at
//@@c

unsigned int lud_wait_read_channel(LUD_CHANNEL* tmc, wait_queue_t *waitqueue_entry)
{
	unsigned long irqstate;
	int retval;
	int send_request = 0;

	if (unlikely(tmc == NULL))
		return 0;

	//printk("Linux-side lud_wait_read_channel() entered\n");

	spin_lock_irqsave(&tmc->lock, irqstate);

	if (tmc->used == 0) {
		/* No data available to read.  Ask the TriMedia to send some. */

		retval = 0;
		if (tmc->pendingrequests == 0) {
			tmc->pendingrequests++;
			send_request = 1;
		}

                if(waitqueue_entry!=NULL) {
                    /* self to filter's wait queue */
                    init_waitqueue_entry(waitqueue_entry,current);
                    add_wait_queue(&tmc->queue,waitqueue_entry);
                }

	}else{
            retval = tmc->used;
        }

	spin_unlock_irqrestore(&tmc->lock, irqstate);

	if (send_request) {
		lud_send_command(tmc->luddev,LUD_SYS_READ_CHANNEL,1,0,tmc->arg,addr_to_token(tmc));
	}
	return retval;
}
//@-node:lud_wait_read_channel
//@+node:lud_unwait_channel
//@+at
//@nonl
// * * Removes the given wait queue entry from the wait queue of this channel.
// * * This function is used after a call to lud_wait_read_channel, normally *
// after the task is woken up. * * This function always returns immediately -
// it does not block. * This function may not be interrupted by signals. * *
// Thread safety: Caller must ensure that they do not attempt multiple *
// simultaneous reads from the same channel. You may simultaneously read *
// from and poll a channel. You must ensure that all operations on the *
// channel have returned before you free the channel. * * @param tmc   The
// channel to read from. * @param waitqueue_entry   The entry to add to this
// channel's wait queue. * @return     0 on success.
//@-at
//@@c

unsigned int lud_unwait_channel(LUD_CHANNEL* tmc, wait_queue_t *waitqueue_entry)
{
	unsigned long irqstate;

	if (unlikely(tmc == NULL))
		return 0;

	spin_lock_irqsave(&tmc->lock, irqstate);

        remove_wait_queue(&tmc->queue,waitqueue_entry);

	spin_unlock_irqrestore(&tmc->lock, irqstate);

        return 0;
}

EXPORT_SYMBOL(lud_write_channel);
EXPORT_SYMBOL(lud_write_channel_ex);
EXPORT_SYMBOL(lud_write_channel_buffered);
EXPORT_SYMBOL(lud_write_channel_buffered_ex);
EXPORT_SYMBOL(lud_write_channel_flush);
EXPORT_SYMBOL(lud_read_channel);
EXPORT_SYMBOL(lud_read_channel_buffered);
EXPORT_SYMBOL(lud_new_write_channel);
EXPORT_SYMBOL(lud_new_read_channel);
EXPORT_SYMBOL(lud_free_channel);
EXPORT_SYMBOL(lud_unwait_channel);
EXPORT_SYMBOL(lud_wait_read_channel);
EXPORT_SYMBOL(lud_poll_read_channel);
EXPORT_SYMBOL(lud_poll_write_channel);

//@-node:lud_unwait_channel
//@-others
//@nonl
//@-node:@file host/channels.c
//@-leo

