//#@+leo-ver=4
//#@+node:@file host/core.h
/* ************************************************************************ *
 * Constants                                                                *
 * ************************************************************************ */

/**
 * The maximum number of TriMedia devices supported by the driver.
 * This can be reduced if memory usage becomes an issue.
 */
#define MAX_LUD_DEV 32

#define LUD_CACHE_ROW_SIZE 64

#define DEF_DPRINTK_LEVEL	1


/**
 * The size of the lud_dbg cyclic buffer.
 */
#define MAX_LUD_DBG	64

/* ************************************************************************ *
 * Types and structures                                                     *
 * ************************************************************************ */

/**
 * The full definition of LUD_CHANNEL.
 * @internal
 * This structure is private to the Luddite core, and external code must
 * only use pointers to it.
 *
 * A one-way communications channel.  May be read or write (but not both).
 * The other end of this channel is an "Xfer" on the target CPU.
 */
struct LUD_CHANNEL_T {
	LUD_DEV_LINK*              luddev; /**< The link to use for communications.
	                                  It is cloned on creation, so this link
	                                  is unique to this channel.  Destroyed
	                                  when the channel is closed.
	                                  Read-only after initialization. */
	volatile int          arg;   /**< The channel ID to pass when communicating
	                                  with the target.
	                                  Read-only after initialization. */
	spinlock_t            lock;  /**< Used to protect the following
	                                  members of this structure. */
	volatile unsigned long addr;  /**< The address of the current buffer, in
	                                  Linux kernel space. */
	volatile int          used;  /**< The number of bytes in the buffer that
	                                  contain data. */
	volatile int          max;   /**< Buffer size.  Only actually used for
	                                  write channels. */
	volatile int          pendingrequests; /**< Number of requests sent to the
	                                  Target that have not yet been answered.
	                                  With the current design, this is always
	                                  0 or 1. */
	volatile int          refcount; /**< Initially 1, free when it reaches zero. */
	wait_queue_head_t     queue;    /**< Queue of processes waiting for a buffer. */
};

/**
 * @internal
 * A Structure representing a single Luddite device
 */
struct LUD_DEV_STRUCT {
	LUD_DEV* luddev;
	struct {
		char*		name;
		char*		opts;
		int		pri;
	} lud_modules[LUDMOD_MAX_MODULES];

	struct LUD_NODE_DEV*  lud_nodes[LUD_MAX_NODES];

	LUD_DEV_LINK* loader;
	LUD_DEV_LINK* system;
	struct semaphore node_sem;	/* mutex on node maintenance */
	struct semaphore mod_sem;	/* mutex on module maintenance */
	struct semaphore sem;
	int	dprintk_level;
};

extern struct class lud_class;
extern struct LUD_DEV_STRUCT lud_devices[];
extern struct file_operations lud_dev_fops;

extern unsigned long lud_remap_addr(LUD_DEV* m,unsigned long addr);
extern void initstr(int);
extern void cleanstr(int);
extern void install_mod(int m,char* n,char* o,int p,int l);
extern int uninstall_mod(int m,char* n);

extern int lud_debug;
extern int clear_target;

extern int register_luddite_device(LUD_DEV*);
extern void unregister_luddite_device(LUD_DEV*);
extern void lud_int_handler(LUD_DEV*);


/*
 * printf(HERE_STRING_FORMAT, HERE_STRING_ARGS) results in a string of the
 * form "file.c:123 (some_func)".   This uses a GCC extension.  Other
 * compilers will return just "file.c:123".
 */
#ifdef __GNUC__
# define HERE_STRING_FORMAT "%s:%d (%s)"
# define HERE_STRING_ARGS   __FILE__, __LINE__, __PRETTY_FUNCTION__
#else
# define HERE_STRING_FORMAT "%s:%d"
# define HERE_STRING_ARGS   __FILE__, __LINE__
#endif

#undef KERN_DEBUG
#define KERN_DEBUG
#define DEB(n,x) do { if (lud_debug >= n) x; } while (0)

//#@-node:@file host/core.h
//#@-leo
