//@+leo-ver=4
//@+node:@file host/node.c
//@@language c
/**
 @file
 */

/**
 * @defgroup group_linux Luddite Linux Side Library
 * These routines are made available by lud-core to all luddite drivers
 * and linux side modules.
 * @{
 */

/* Copyright Philips Research Labs, Redhill, UK. 2001
 *
 * Prototype Software
 *
 * Contact: Tim Everett (tim.everett@philips.com)
 *
 */

/***** Includes *****/

//@<< lud-base #includes >>
//@+node:<< lud-base #includes >>
#ifdef CONFIG_SMP
#define __SMP__
#endif

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/page.h>
#include <asm/semaphore.h>
#include <asm/atomic.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>

#include <linux/module.h>

#include <linux/version.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/kmod.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#include <linux/device.h>
#include <linux/cdev.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18))
#include <linux/compiler.h>
#else
#define likely(x) (x)
#define unlikely(x) (x)
#endif /* LINUX_VERSION_CODE */

#include <stdarg.h>


#include "compat.h"
#include "lud_mmio.h"
#include "luddite-host.h"
#include "luddite/version.h"
#include "luddite/os.h"
#include "exported.h"

#include "core.h"
#include "procfs.h"
#include "node.h"
#include "helpers.h"
#include "channels.h"
//@-node:<< lud-base #includes >>
//@nl
//@<< lud-node declarations >>
//@+node:<< lud-node declarations >>

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#define INC_LUD_MOD_USE_COUNT(tmlink_ponter) do {} while (0);
#define DEC_LUD_MOD_USE_COUNT(tmlink_ponter) do {} while (0);
#else
#define INC_LUD_MOD_USE_COUNT(tmlink_ponter) \
__MOD_INC_USE_COUNT(lud_devices[(tmlink_ponter)->minor].luddev->driver_module)
#define DEC_LUD_MOD_USE_COUNT(tmlink_ponter) \
__MOD_DEC_USE_COUNT(lud_devices[(tmlink_ponter)->minor].luddev->driver_module)
#endif

#ifdef CONFIG_SYSFS
/*
* Temporary minor number allocator
*/

#define MAX_MINOR_NUMS	256

static unsigned char minor_numbers[MAX_MINOR_NUMS-MAX_LUD_DEV];
//@-node:<< lud-node declarations >>
//@nl
//@+others
//@+node:next_free_minor

#define free_dev_minor(d)	minor_numbers[MINOR(d)-MAX_LUD_DEV] = 0
static int next_free_minor(void)
{
	int i;
	for (i=0;i<MAX_MINOR_NUMS-MAX_LUD_DEV;i++) {
		if (minor_numbers[i]==0) {
			minor_numbers[i] = 1;
			return (i+MAX_LUD_DEV);
		}
	}
	printk("lud: no minor numbers left\n");
	return -1;
}
//@-node:next_free_minor
//@+node:init_minor_numbers
void init_minor_numbers(void)
{
	int i;
	for (i=0;i<MAX_MINOR_NUMS-MAX_LUD_DEV;i++) minor_numbers[i]=0;
}
//@-node:init_minor_numbers
//@+node:lud_node_open

#endif
/* ************************************************************************ *
 * Node file operations                                                     *
 * ************************************************************************ */

static int lud_node_open(struct inode * inode,
                         struct file  * file)
{
#ifdef CONFIG_SYSFS
	struct LUD_NODE_DEV* node = container_of(inode->i_cdev,struct LUD_NODE_DEV,cdev);
#else
	struct LUD_NODE_DEV* node = (struct LUD_NODE_DEV*)file->private_data;
#endif
	int requested_mode;

	if (unlikely(node->dev.dev->state != LUD_LUDDITE)) {
		return -EIO;
	}

#ifdef CONFIG_SYSFS
	file->private_data = (void*)node;
#endif

	/* MOD_INC_USE_COUNT; - already done due to SET_MODULE_OWNER(fops) */
	INC_LUD_MOD_USE_COUNT(&(node->dev));

	if (unlikely(node->flags & LUD_NODE_SERVER)) {
		/* Server node.  Must ask for real node. */
		int connection_handle;
		LUD_DEV_LINK * server_link = lud_clone_link(&node->dev);
		if (unlikely(server_link == NULL)) {
			DEC_LUD_MOD_USE_COUNT(&(node->dev));
			return -ENOMEM;
		}

		connection_handle = lud_send_request(server_link, LUD_SYS_NEW_CONNECTION, 0, 0, 0, 0);
		lud_free_link(server_link);
		if (unlikely(connection_handle == -1)) {
			printk(KERN_ERR "Interrupted system call recovery not yet implemented in %s:%d\n", __FILE__, __LINE__);
			DEC_LUD_MOD_USE_COUNT(&(node->dev));
			return -EINTR;
		}
		if ( unlikely(connection_handle <= 0)
		  || unlikely(connection_handle >= LUD_MAX_NODES)) {
			if (likely(connection_handle == 0)) {
				printk(KERN_INFO "Target refused server node connection\n");
			} else {
				printk(KERN_ERR "Target returned invalid node on server connect request\n");
			}
			DEC_LUD_MOD_USE_COUNT(&(node->dev));
			return -EIO;
		}

		node = lud_devices[node->dev.minor].lud_nodes[connection_handle];

		if (unlikely(!node)) {
			/* No such node */
			printk(KERN_ERR "Target returned non-existant node on server connect request\n");
			DEC_LUD_MOD_USE_COUNT(&(node->dev));
			return -EIO;
		}

		/* Redirect to real node - 'node' has already been changed. */
		file->private_data = (void *)node;
	}

	/* Atomically test the appropriate bits to see if this node is
	 * already open for (any of) the desired mode(s), and fail if so.
	 * If not, set the appropriate bits.
	 * We use a spinlock to prevent a classical test+set race condition.
	 *
	 * This test is optional and may be disabled using the
	 * LUD_NODE_MULTIPLE_READERS and/or LUD_NODE_MULTIPLE_WRITERS
	 * flags when the node is created.
	 */

	requested_mode = file->f_mode & (FMODE_READ|FMODE_WRITE);
	if (node->flags & LUD_NODE_MULTIPLE_READERS)
		requested_mode &= ~FMODE_READ;
	if (node->flags & LUD_NODE_MULTIPLE_WRITERS)
		requested_mode &= ~FMODE_WRITE;

	spin_lock(&node->state_lock);
	if (unlikely(requested_mode & node->state)) {
		/* already open for one of the operations */
		spin_unlock(&node->state_lock);
		DEC_LUD_MOD_USE_COUNT(&(node->dev));
		return -EBUSY;
	}
	node->state |= requested_mode;
	spin_unlock(&node->state_lock);

	if (file->f_mode & FMODE_READ) {
		if (node->flags & LUD_NODE_TARGET_HOST) {
			node->read = lud_new_read_channel(&node->dev);
			if (unlikely(node->read == NULL)) {
				DEC_LUD_MOD_USE_COUNT(&(node->dev));
				return -ENOMEM;
			}
		}
	}

	if (file->f_mode & FMODE_WRITE) {
		if (node->flags & LUD_NODE_HOST_TARGET) {
			node->write = lud_new_write_channel(&node->dev);
			if (unlikely(node->write == NULL)) {
				DEC_LUD_MOD_USE_COUNT(&(node->dev));
				return -ENOMEM;
			}
		}
	}

	return 0;
}
//@-node:lud_node_open
//@+node:lud_node_release

static int lud_node_release(struct inode * inode,
                            struct file  * file)
{
	struct LUD_NODE_DEV* node = (struct LUD_NODE_DEV*)file->private_data;

	if (unlikely(!node)) {
		return -EIO;
	}

	if (file->f_mode & FMODE_READ) {
		if (node->read) {
			lud_free_channel(node->read);
			node->read = 0;
		}
	}
	if (file->f_mode & FMODE_WRITE) {
		if (node->write) {
			lud_write_channel_flush(node->write);
			lud_free_channel(node->write);
			node->write = 0;
		}
	}

	spin_lock(&node->state_lock);
	node->state &= ~(file->f_mode & (FMODE_READ|FMODE_WRITE));
	spin_unlock(&node->state_lock);

	DEC_LUD_MOD_USE_COUNT(&(node->dev));
	/* MOD_DEC_USE_COUNT; - will be done automatically due to SET_MODULE_OWNER(fops) */

	return 0;
}
//@-node:lud_node_release
//@+node:lud_node_read

static ssize_t lud_node_read(struct file * file,
                             char * buffer,
                             size_t count,
                             loff_t *ppos)
{
	struct LUD_NODE_DEV* node = (struct LUD_NODE_DEV*)file->private_data;

	if (unlikely(!node)) {
		return -EIO;
	}

	if (unlikely(node->dev.dev->state != LUD_LUDDITE)) {
		return -EIO;
	}

	return lud_generic_read(node->read, &node->read_sema, file, buffer, count, ppos);
}
//@-node:lud_node_read
//@+node:lud_node_write

static ssize_t lud_node_write(struct file * file,
                              const char * buffer,
                              size_t count,
                              loff_t *ppos)
{
	struct LUD_NODE_DEV* node = (struct LUD_NODE_DEV*)file->private_data;

	if (unlikely(!node)) {
		return -EIO;
	}

	if (unlikely(node->dev.dev->state != LUD_LUDDITE)) {
		return -EIO;
	}


	return lud_generic_write((node->flags & LUD_NODE_AUTO_FLUSH), node->write, &node->write_sema, file, buffer, count, ppos);
}
//@-node:lud_node_write
//@+node:lud_node_poll
//@+at
//@nonl
// * * poll() or select() over the device.
//@-at
//@@c

static unsigned int lud_node_poll(struct file * file, poll_table * wait)
{
	struct LUD_NODE_DEV* node = (struct LUD_NODE_DEV*)file->private_data;

	if (unlikely(!node)) {
		return -EIO;
	}

	if (unlikely(node->dev.dev->state != LUD_LUDDITE)) {
		return -EIO;
	}

	return lud_poll_write_channel(node->write, file, wait)
	     | lud_poll_read_channel (node->read, file, wait);
}
//@-node:lud_node_poll
//@+node:lud_node_ioctl

static int lud_node_ioctl(struct inode * inode,
	struct file * file,
	unsigned int cmd,
	unsigned long arg)
{
	struct LUD_NODE_DEV* node = (struct LUD_NODE_DEV*)file->private_data;

	if (unlikely(!node)) {
		return -EBADF ;		/* more suitable than -EIO? */
	}

	if (unlikely(node->dev.dev->state != LUD_LUDDITE)) {
		return -EIO;
	}

	if (unlikely(cmd == LUD_SYS_FLUSH)) {
		int err = 0 ;
		if (likely(!(node->flags & LUD_NODE_AUTO_FLUSH))) {
			if (unlikely((file->f_flags & O_NONBLOCK) == O_NONBLOCK)) {
				if (unlikely(down_trylock(&node->write_sema))) {
					return -EAGAIN;
				}
			} else {
				if (unlikely(down_interruptible(&node->write_sema))) {
					return -ERESTARTSYS;
				}
			}

			err = lud_write_channel_flush(node->write);

			up(&node->write_sema);
		}
		return err ;
	}

	return lud_generic_ioctl(&node->dev, cmd, arg);
}
//@-node:lud_node_ioctl
//@+node:lud_generic_read

struct file_operations lud_node_fops = {
read: 	lud_node_read,
write: 	lud_node_write,
ioctl: 	lud_node_ioctl,
open: 	lud_node_open,
release:lud_node_release,
llseek:	lud_generic_llseek,
poll:	lud_node_poll,
};


/* ************************************************************************ *
 * Generic file operations                                                  *
 * ************************************************************************ */

/**
 * Generic implementation of file read() function.
 *
 * @param xfer   The Xfer to use.  If NULL, -EIO is returned.
 * @param sema   The semaphore that prevents simultaneous reads.  If NULL, no
 *               semaphore is used.
 * @param file   As passed to the driver by the Linux kernel.
 * @param buffer As passed to the driver by the Linux kernel.
 * @param count  As passed to the driver by the Linux kernel.
 * @param ppos   As passed to the driver by the Linux kernel.
 * @return       Suitable return value to be passed back to the Linux kernel.
 */
ssize_t lud_generic_read(LUD_CHANNEL * xfer,
                        struct semaphore * sema,
                        struct file * file,
                        char * buffer,
                        size_t count,
                        loff_t *ppos)
{
	int result;
	int nonblock = ((file->f_flags & O_NONBLOCK) == O_NONBLOCK);

	if (unlikely(xfer == NULL)) {
		return -EIO;
	}

	if (unlikely(ppos != &file->f_pos)) {
		/* Disallow pread() - this is a pipe, not a file, so it's not seekable. */
		/* return -ESPIPE; */
	}

	if (unlikely(!access_ok(VERIFY_WRITE, buffer, count))) {
		return -EFAULT;
	}

	if (likely(sema != NULL)) {
		if (unlikely(nonblock)) {
			if (unlikely(down_trylock(sema))) {
				return -EAGAIN;
			}
		} else {
			if (unlikely(down_interruptible(sema))) {
				return -ERESTARTSYS;
			}
		}
	}
	result = lud_read_channel(xfer, buffer, count, 0, nonblock);

	if (likely(result > 0)) {
		/* success, update file position */
		*ppos += result;
	}

	if (likely(sema != NULL)) {
		up(sema);
	}

	return result;
}
//@-node:lud_generic_read
//@+node:lud_generic_write
//@+at
//@nonl
// * * Generic implementation of file write() function. * * @param flush If
// nonzero, the data is always flushed after the write. * @param xfer  The
// Xfer to use. If NULL, -EIO is returned. * @param sema  The semaphore that
// prevents simultaneous writes. If NULL, no *        semaphore is used. *
// @param file  As passed to the driver by the Linux kernel. * @param buffer
// As passed to the driver by the Linux kernel. * @param count As passed to
// the driver by the Linux kernel. * @param ppos  As passed to the driver by
// the Linux kernel. * @return    Suitable return value to be passed back to
// the Linux kernel.
//@-at
//@@c

ssize_t lud_generic_write(int flush,
                         LUD_CHANNEL * xfer,
                         struct semaphore * sema,
                         struct file * file,
                         const char * buffer,
                         size_t count,
                         loff_t *ppos)
{
	int result;
	int nonblock = ((file->f_flags & O_NONBLOCK) == O_NONBLOCK);

	if (unlikely(xfer == NULL)) {
		return -EIO;
	}

	if (unlikely(ppos != &file->f_pos)) {
		/* Disallow pwrite() - this is a pipe, not a file, so it's not seekable. */
		/*return -ESPIPE; */
	}

	if (unlikely(!access_ok(VERIFY_READ, buffer, count))) {
		return -EFAULT;
	}

	if (likely(sema != NULL)) {
		if (unlikely(nonblock)) {
			if (unlikely(down_trylock(sema))) {
				return -EAGAIN;
			}
		} else {
			if (unlikely(down_interruptible(sema))) {
				return -ERESTARTSYS;
			}
		}
	}

	result = lud_write_channel_ex(xfer, buffer, count, 0, nonblock, flush);

	if (likely(result > 0)) {
		/* success, update file position */
		*ppos += result;
	}

	if (likely(sema != NULL)) {
		up(sema);
	}

	return result;
}
//@-node:lud_generic_write
//@+node:lud_generic_ioctl
//@+at
//@nonl
// * * Generic implementation of file ioctl() function. * * @param link  The
// Link to use. If NULL, -EIO is returned. * @param cmd  The Linux ioctl
// number. * @param arg  The Linux ioctl argument. * @return    Suitable
// return value to be passed back to the Linux kernel.
//@-at
//@@c

int lud_generic_ioctl(LUD_DEV_LINK *link,
                     unsigned int cmd,
                     unsigned long arg)
{
	unsigned int need_ack = _IOC_DIR(cmd)&_IOC_READ;
	unsigned int c_type =   _IOC_TYPE(cmd);
	unsigned int c_index = _IOC_NR(cmd);
	unsigned int size = _IOC_SIZE(cmd);
	unsigned int lud_cmd;
	unsigned int arg0 = 0;
	unsigned int arg1 = 0;
	unsigned int arg2 = 0;
	unsigned int arg3 = 0;
	unsigned int *iarg = (unsigned int *)arg;

	if (unlikely(link == NULL)) {
		return -EIO;
	}

	if (unlikely(size > 16) || unlikely((size & 3) != 0)) {
		/* Invalid IOCTL.  size must be 0, 4, 8, 12, or 16. */
		return -ENOTTY;
	}

	if (need_ack) {
		lud_cmd = LUD_IOCTL_REQ(c_type,c_index);

		if (likely(size != 0) && unlikely(!access_ok(VERIFY_WRITE,iarg,size))) {
			return -EFAULT;
		}
	} else {
		lud_cmd = LUD_IOCTL_CMD(c_type,c_index);

		if (likely(size != 0) && unlikely(!access_ok(VERIFY_READ,iarg,size))) {
			return -EFAULT;
		}
	}

	switch (size) {
	case 0:
		arg0 = arg;
		break;
	case 4:
		if (unlikely(__get_user(arg0,iarg)))
			return -EFAULT;
		break;
	case 8:
		if (unlikely(__get_user(arg0,iarg))
		 || unlikely(__get_user(arg1,iarg+1)))
			return -EFAULT;
		break;
	case 12:
		if (unlikely(__get_user(arg0,iarg))
		 || unlikely(__get_user(arg1,iarg+1))
		 || unlikely(__get_user(arg2,iarg+2)))
			return -EFAULT;
		break;
	case 16:
		if (unlikely(__get_user(arg0,iarg))
		 || unlikely(__get_user(arg1,iarg+1))
		 || unlikely(__get_user(arg2,iarg+2))
		 || unlikely(__get_user(arg3,iarg+3)))
			return -EFAULT;
		break;
	default:
		BUG();
	}

	if (need_ack) {
		int ret = lud_send_request(link,lud_cmd,arg0,arg1,arg2,arg3);
		switch (size) {
		case 0:
			return ret;
		case 4:
			if (unlikely(__put_user(link->retval0,iarg)))
				return -EFAULT;
			break;
		case 8:
			if (unlikely(__put_user(link->retval0,iarg))
			 || unlikely(__put_user(link->retval1,iarg+1)))
				return -EFAULT;
			break;
		case 12:
			if (unlikely(__put_user(link->retval0,iarg))
			 || unlikely(__put_user(link->retval1,iarg+1))
			 || unlikely(__put_user(link->retval2,iarg+2)))
				return -EFAULT;
			break;
		case 16:
			if (unlikely(__put_user(link->retval0,iarg))
			 || unlikely(__put_user(link->retval1,iarg+1))
			 || unlikely(__put_user(link->retval2,iarg+2))
			 || unlikely(__put_user(link->retval3,iarg+3)))
				return -EFAULT;
			break;
		default:
			BUG();
		}
	} else {
		lud_send_command(link,lud_cmd,arg0,arg1,arg2,arg3);
	}

	return 0;
}
//@-node:lud_generic_ioctl
//@+node:lud_generic_llseek
//@+at
//@nonl
// * * Handle seeking. Luddite nodes do not support this, so this method
// always fails. * * @param file Ignored. * @param off Ignored. * @param
// whence Ignored. * @return Error code: cannot seek a pipe. (-ESPIPE).
//@-at
//@@c

loff_t lud_generic_llseek(struct file * file, loff_t off, int whence)
{
	return -ESPIPE;
}
//@-node:lud_generic_llseek
//@+node:lud_node_restore_fops

void lud_node_restore_fops(struct LUD_NODE_DEV* node)
{
#ifdef CONFIG_SYSFS
	node->cdev.ops = &lud_node_fops;
#else
	if (node->dev_h != NULL) {
		/* remove existing node */
		if (node->read) {
			lud_free_channel(node->read);
			node->read = NULL;
		}
		if (node->write) {
			lud_free_channel(node->write);
			node->write = NULL;
		}
		if (node->dev_h) {
			devfs_unregister(node->dev_h);
			node->dev_h = NULL;
		}

		/* create new entry */
		node->dev_h = devfs_register(node->dev.dev->devfs_directory,
				node->name,
			        DEVFS_FL_AUTO_DEVNUM,
			        0, 0,
			        S_IFCHR | S_IRUGO | S_IWUGO,
			        &lud_node_fops, node);
	}
#endif
}
//@-node:lud_node_restore_fops
//@+node:lud_node_new_fops

#ifdef CONFIG_SYSFS
unsigned long*
#else
devfs_handle_t
#endif
lud_node_new_fops(struct LUD_NODE_DEV* node,
				struct file_operations* fops)
{
#ifdef CONFIG_SYSFS
	node->cdev.ops = fops;
#else
	if (node->dev_h != NULL) {
		/* remove existing node */
		if (node->read) {
			lud_free_channel(node->read);
			node->read = NULL;
		}
		if (node->write) {
			lud_free_channel(node->write);
			node->write = NULL;
		}
		if (node->dev_h) {
			devfs_unregister(node->dev_h);
			node->dev_h = NULL;
		}

		/* create new entry */
		node->dev_h = devfs_register(node->dev.dev->devfs_directory,
				node->name,
			        DEVFS_FL_AUTO_DEVNUM,
			        0, 0,
			        S_IFCHR | S_IRUGO | S_IWUGO,
			        fops, node);
		return node->dev_h;
	}
#endif
	return NULL;
}
//@-node:lud_node_new_fops
//@+node:node_dev_release
//@+at
//@nonl
// ************************************************************************ *
// * Node creation/deletion                          * *
// ************************************************************************
//@-at
//@@c

#ifdef CONFIG_SYSFS
static void node_dev_release(struct device* dev)
{
}
//@-node:node_dev_release
//@+node:show_ldev_luddite

static ssize_t show_ldev_luddite(struct class_device *dev, char *buf)
{
        char *str = buf;
        str += sprintf(str,"1");
        return (str - buf);
}
//@-node:show_ldev_luddite
//@+node:show_ldev_name
static CLASS_DEVICE_ATTR(luddite,S_IRUGO,show_ldev_luddite,NULL);
static ssize_t show_ldev_name(struct class_device *dev, char *buf)
{
    char *str = buf;
	struct LUD_NODE_DEV* t = dev->class_data;
    str += sprintf(str,"%s\n",t->name);
    return (str - buf);
}
//@-node:show_ldev_name
//@+node:show_ldev_dspno
static CLASS_DEVICE_ATTR(name,S_IRUGO,show_ldev_name,NULL);
static ssize_t show_ldev_dspno(struct class_device *dev, char *buf)
{
    char *str = buf;
	struct LUD_NODE_DEV* t = dev->class_data;
    str += sprintf(str,"%d\n",t->dev.minor);
    return (str - buf);
}
//@-node:show_ldev_dspno
//@+node:show_ldev_dev
static CLASS_DEVICE_ATTR(dspno,S_IRUGO,show_ldev_dspno,NULL);

static ssize_t show_ldev_dev(struct class_device *dev, char *buf)
{
    char *str = buf;
	struct LUD_NODE_DEV* t = dev->class_data;
    str += sprintf(str,"%d:%d\n", MAJOR(t->cdev.dev),MINOR(t->cdev.dev));
    return (str - buf);
}
//@-node:show_ldev_dev
//@+node:new_node
static CLASS_DEVICE_ATTR(dev,S_IRUGO,show_ldev_dev,NULL);
#endif

void new_node(int minor,int handle,char* name,int flags)
{
	struct LUD_NODE_DEV * node ;

	if (unlikely((minor < 0) || (minor >= MAX_LUD_DEV)))
		return ;
	if (unlikely((handle < 0) || (handle >= LUD_MAX_NODES)))
		return ;

	node = lud_devices[minor].lud_nodes[handle];

	if (node)
		return;

	down(&lud_devices[minor].node_sem);

	node = (struct LUD_NODE_DEV*)kmalloc(sizeof(struct LUD_NODE_DEV),GFP_KERNEL);
	if (unlikely(!node)) {
		return ;
	}
	memset(node,0,sizeof(struct LUD_NODE_DEV));

	lud_devices[minor].lud_nodes[handle] = node;

	node->exists = 1;

	strncpy(node->name, name, sizeof(node->name) - 1);
	node->name[sizeof(node->name) - 1] = '\0';

	node->dev_h = NULL;
	node->read = NULL;
	node->write = NULL;
	node->state = 0;
	node->flags = flags;

	spin_lock_init(&node->state_lock);

	sema_init(&node->read_sema,  1);
	sema_init(&node->write_sema, 1);

	node->dev.major = handle;
	node->dev.minor = minor;
	node->dev.dev = lud_devices[minor].luddev;
	node->private_data = NULL;
	atomic_set(&node->dev.freecounter, 1);
	init_waitqueue_head(&node->dev.q);

	if ( ((flags & (LUD_NODE_TARGET_HOST | LUD_NODE_HOST_TARGET | LUD_NODE_HOST_IOCTL)) != 0)
	  && ((flags & LUD_NODE_INVISIBLE) == 0) ) {
#ifdef CONFIG_SYSFS
		dev_t devt = MKDEV(LUD_MAJOR, next_free_minor());
		node->dev_h = (unsigned long*)1;
		cdev_init(&node->cdev,&lud_node_fops);
		node->cdev.owner = THIS_MODULE;
		kobject_set_name(&node->cdev.kobj,node->name);
		if (cdev_add(&node->cdev, devt, 1)) {
			printk("Error adding cdev for %s\n",node->name);
		}

		/* set up bus reg */
		memset(&node->ldev,0,sizeof(struct device));
		node->ldev.bus = NULL;
		node->ldev.parent = &node->dev.dev->ldev;
		node->ldev.release = node_dev_release;
		node->ldev.devt = devt;
		snprintf(node->ldev.bus_id, BUS_ID_SIZE, "%s",node->name);
		device_register(&node->ldev);
		node->classdev = class_device_create(&lud_class, NULL, devt, &node->ldev, "%d:%s", minor, node->name);
		node->classdev->class_data = node;
		class_device_create_file(node->classdev,&class_device_attr_dev);
		class_device_create_file(node->classdev,&class_device_attr_dspno);
		class_device_create_file(node->classdev,&class_device_attr_name);
		class_device_create_file(node->classdev,&class_device_attr_luddite);
#else
		node->dev_h =
			devfs_register(lud_devices[minor].luddev->devfs_directory, node->name,
			               DEVFS_FL_AUTO_DEVNUM,
			               0, 0,
			               S_IFCHR | S_IRUGO | S_IWUGO,
			               &lud_node_fops, lud_devices[minor].lud_nodes[handle]);
#endif
	}
	up(&lud_devices[minor].node_sem);
}
//@-node:new_node
//@+node:del_node

void del_node(int minor,int handle)
{
	struct LUD_NODE_DEV * node ;

	if (unlikely((minor < 0) || (minor >= MAX_LUD_DEV)))
		return ;
	if (unlikely((handle < 0) || (handle >= LUD_MAX_NODES)))
		return ;

	node = lud_devices[minor].lud_nodes[handle];

	if (node) {
		down(&lud_devices[minor].node_sem);

		lud_devices[minor].lud_nodes[handle] = NULL;
		node->exists = 0;

		if (node->handler) {
			printk(KERN_ERR "lud-core warning: Deleting node %s which is in use by handler <%p>.\n",
			       node->name, (void *)node->handler);
			node->handler = NULL;
			DEC_LUD_MOD_USE_COUNT(&node->dev);
		}

		node->name[0] = '\0';

		if (node->dev_h) {
			if (node->read) {
				lud_free_channel(node->read);
				node->read = NULL;
			}
			if (node->write) {
				lud_free_channel(node->write);
				node->write = NULL;
			}
			if (node->dev_h) {
#ifdef CONFIG_SYSFS
/* Not safe to do this here!!! but for now... */
 		 		class_device_remove_file(node->classdev,&class_device_attr_luddite);
 		 		class_device_remove_file(node->classdev,&class_device_attr_dspno);
 		 		class_device_remove_file(node->classdev,&class_device_attr_name);
 		 		class_device_remove_file(node->classdev,&class_device_attr_dev);
 		 		class_device_unregister(node->classdev);
				device_unregister(&node->ldev);
				free_dev_minor(node->cdev.dev);
				cdev_del(&node->cdev);
#else
				devfs_unregister(node->dev_h);
#endif
				node->dev_h = NULL;
			}
		}

		kfree(node);

		up(&lud_devices[minor].node_sem);
	}
}
//@-node:del_node
//@-others
EXPORT_SYMBOL(lud_node_new_fops);
EXPORT_SYMBOL(lud_node_restore_fops);
EXPORT_SYMBOL(lud_generic_read);
EXPORT_SYMBOL(lud_generic_write);
EXPORT_SYMBOL(lud_generic_ioctl);
EXPORT_SYMBOL(lud_generic_llseek);
/**
 * @}
 */
//@-node:@file host/node.c
//@-leo
