/*
   Copyright (C) 2005 Koninklijke Philips Electronics N.V.,
   All Rights Reserved.

Rev Date       Author      Comments
--------------------------------------------------------------------------------
  1 20051130   burningh    Original
  2 20051206   burningh    Support multiple readers in non-wrapping mode
  3 20051207   burningh    Allow a blocked reader to exit when all writers exit
  4 20051208   burningh    Add ioctl to control wrap mode of DSP pipe
  5 20060125   burningh    Fix issues raised by ASL
  6 20060314   burningh    Implement poll file_op to allow select, poll & epoll
6.1.1 20060908 laird       Move to Linux 2.6.17.7
*/
/** DspPipe driver for PNX8550.
 * Acts like a pipe, except for the following:
 *   - has a wrapping / non wrapping mode (whether the write pointer can wrap around or not)
 *   - in non-wrapping mode, sequential readers always begin from the start of
 *      the buffer
 *   - in wrapping mode, it acts like a normal pipe
 *
 * This is needed to allow detection graphs & DirectFB DataBuffer reads to work
 * as expected with streamed data. */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/proc_fs.h>
#include <linux/wait.h>
#include <linux/poll.h>
#include <asm/semaphore.h>

#include <phStbDspPipe.h>

#include <asm/mach-pnx8550/glb.h>
#include <phStbProc_BaseProcSystem.h>

/******************************************************************************
* LOCAL MACROS                                                                *
*******************************************************************************/

#define THIS_MODULE_DESCRIPTION "DSP Pipe driver"

/* The default number of devices we support */
#define NR_DEVS 2

/* A debug macro */
#define DEBUG(x) if (debug) printk x

/* Module set up */
#define MY_LICENSE "GPL"
MODULE_LICENSE(MY_LICENSE);

MODULE_DESCRIPTION(THIS_MODULE_DESCRIPTION);
MODULE_AUTHOR("Neil Burningham");

#define TRUE 1
#define FALSE 0

/******************************************************************************
* STATIC FUNCTION PROTOTYPES                                                  *
*******************************************************************************/

static int phStbDspPipe_open(struct inode *inode, struct file *file);
static int phStbDspPipe_release(struct inode *inode, struct file *file);
static ssize_t phStbDspPipe_read(struct file *file, char __user *buf, size_t count, loff_t *pos);
static ssize_t phStbDspPipe_write(struct file *file, const char __user *buf, size_t count, loff_t *pos);
static int phStbDspPipe_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static unsigned int phStbDspPipe_poll(struct file *file, poll_table *wait);

/******************************************************************************
* LOCAL TYPEDEFS                                                              *
*******************************************************************************/

/* Data about the buffer held by a device */
struct phStbDspPipe_Buffer_t
{
    struct semaphore sem; /* The mutex used to protect this data */
    char *pStart; /* Pointer to the start of the buffer */
    char *pEnd; /* Pointer to the end of the buffer */
    ssize_t bufferSize; /* The size of the buffer */
    char *pRead, *pWrite; /* The read and write pointers */
    wait_queue_head_t writerQ, readerQ; /* The reader & writer sleep queues */
    int noReaders, noWriters; /* The number of readers & writers */
    int writerReleased; /* Whether the writer has released the pipe */
    int wrapAllowed; /* Whether overwriting of data is allowed or not */
};

/* Data held by a driver instance */
struct phStbDspPipe_t
{
    dev_t devRegion; /* The start of the reserved device number region */
    struct cdev *cdev; /* The array of character devices */
    struct class *sysClass; /* The class under /sys/class */
    unsigned int major; /* Major number of our device, obtained from devRegion and held for convenience */
    struct proc_dir_entry *procDir; /* Our directory under /proc */
    struct phStbDspPipe_Buffer_t *buffers; /* The buffer descriptors for each device */
};

/******************************************************************************
* EXPORTED DATA                                                               *
*******************************************************************************/

/******************************************************************************
* LOCAL DATA                                                                  *
*******************************************************************************/

/* The driver instance data */
static struct phStbDspPipe_t gInstance;

/* The file operations for the character devices */
static struct file_operations phStbDspPipe_fops =
{
    .owner      = THIS_MODULE,
    .open       = phStbDspPipe_open,
    .release    = phStbDspPipe_release,
    .read       = phStbDspPipe_read,
    .write      = phStbDspPipe_write,
    .poll       = phStbDspPipe_poll,
    .ioctl      = phStbDspPipe_ioctl
};

/* Some of these parameters can be modified while the module is installed.
 * buffer_size affects the size of dsppipe device buffers when they are created (ref count goes from 0 to 1)
 * wrap_allowed affects the wrapping status of a buffer when it is created (ref count goes from 0 to 1)
 * debug takes immediate effect */
static int wrap_allowed = 0;
static uint buffer_size = PAGE_SIZE * 8;
static int debug = 0;
static uint nr_devs = NR_DEVS;

module_param(wrap_allowed,int,S_IRUGO|S_IWUSR);
MODULE_PARM_DESC(wrap_allowed, "Affects the wrapping status of a device when it is created (ref count goes from 0 to 1)."
                 " 1 means wrap (fifo behaviour), 0 means no wrap (flat file behaviour).");

module_param(buffer_size,uint,S_IRUGO|S_IWUSR);
MODULE_PARM_DESC(buffer_size, "Affects the size of a device when it is created (ref count goes from 0 to 1)");

module_param(debug,int,S_IRUGO|S_IWUSR);
MODULE_PARM_DESC(debug, "Determines whether debugging prints are emitted. 0 means no debug, 1 means debug.");

module_param(nr_devs,uint,S_IRUGO);
MODULE_PARM_DESC(nr_devs, "Affects the number of devices this driver supports.");

/******************************************************************************
* FUNCTION IMPLEMENTATION                                                     *
*******************************************************************************/

/* Open the device - set things up if necessary */
static int phStbDspPipe_open(struct inode *inode, struct file *file)
{
    int err = -ENODEV;

    DEBUG((KERN_INFO "DspPipe_open: inode %p, file %p\n", inode, file));

    if (iminor(inode) < nr_devs)
    {
        struct phStbDspPipe_Buffer_t *buffer = &gInstance.buffers[iminor(inode)];

        if (down_interruptible(&buffer->sem))
            return -ERESTARTSYS;

        /* Ensure that we have a maximum of one reader & writer per device */
        if (file->f_mode & FMODE_READ)
        {
            buffer->noReaders++;
        }
        if (file->f_mode & FMODE_WRITE)
        {
            if (buffer->noWriters > 0)
            {
                up(&buffer->sem);
                return -EBUSY;
            }
            buffer->noWriters++;
        }

        file->private_data = buffer;

        if (buffer->pStart == NULL)
        {
            buffer->bufferSize = buffer_size;
            buffer->pStart = kmalloc(buffer->bufferSize, GFP_KERNEL);
            if (buffer->pStart == NULL)
                return -ENOMEM;
            buffer->pEnd = buffer->pStart + buffer->bufferSize;
            buffer->pRead = buffer->pWrite = buffer->pStart;

            buffer->writerReleased = FALSE;

            buffer->wrapAllowed = wrap_allowed;
        }

        up(&buffer->sem);

        /* Needed in case the wraparound parameter has changed to non-zero,
         * meaning that writers are now allowed to write up to the read pointer */
        wake_up_interruptible(&buffer->writerQ);

        err = 0;
    }

    return err;
}


/* Release the device */
static int phStbDspPipe_release(struct inode *inode, struct file *file)
{
    int err = -ENODEV;

    DEBUG((KERN_INFO "DspPipe_release: inode %p, file %p\n", inode, file));

    if (iminor(inode) < nr_devs)
    {
        struct phStbDspPipe_Buffer_t *buffer = &gInstance.buffers[iminor(inode)];

        if (down_interruptible(&buffer->sem))
            return -ERESTARTSYS;

        if (file->f_mode & FMODE_READ)
            buffer->noReaders--;

        if (file->f_mode & FMODE_WRITE)
            buffer->noWriters--;

        if ((buffer->noReaders == 0) && (buffer->noWriters == 0))
        {
            DEBUG((KERN_INFO "DspPipe_release: freeing unused buffer\n"));
            kfree(buffer->pStart);
            buffer->pStart = buffer->pEnd = buffer->pRead = buffer->pWrite = NULL;
        }

        /* If wrapping is off, and this file pointer is for reading, reset the read pointer. */
        else if ((file->f_mode & FMODE_READ) && !buffer->wrapAllowed)
        {
            DEBUG((KERN_INFO "DspPipe_release: resetting read pointer to beginning of buffer.\n"));
            buffer->pRead = buffer->pStart;
        }

        up(&buffer->sem);

        /* If this is a writer exiting, wake up readers so they can exit. */
        if (file->f_mode & FMODE_WRITE)
        {
            buffer->writerReleased = TRUE;
            wake_up_interruptible(&buffer->readerQ);
        }

        err = 0;
    }

    return err;
}


/* Read data from the current position in the file */
static ssize_t phStbDspPipe_read(struct file *file, char __user *buf, size_t count, loff_t *pos)
{
    struct phStbDspPipe_Buffer_t *buffer = (struct phStbDspPipe_Buffer_t*)file->private_data;

    DEBUG((KERN_INFO "DspPipe_read: file %p, buf %p, count %d, start %p, read %p, write %p, end %p\n",
           file, buf, count, buffer->pStart, buffer->pRead, buffer->pWrite, buffer->pEnd));

    if (down_interruptible(&buffer->sem))
        return -ERESTARTSYS;

    /* Handle the situation where there is no data to read - block or return -EAGAIN.
     * Ensure that we don't block at the end of the "flat file" in non-wrapping mode. */
    while (!buffer->writerReleased &&
           ((buffer->wrapAllowed && (buffer->pRead == buffer->pWrite)) ||
            (!buffer->wrapAllowed && (buffer->pRead == buffer->pWrite) && (buffer->pRead < (buffer->pEnd - 1)))))
    {
        up(&buffer->sem);

        if (file->f_flags & O_NONBLOCK)
            return -EAGAIN;

        if (wait_event_interruptible(buffer->readerQ, ((buffer->pRead != buffer->pWrite) || buffer->writerReleased)))
            return -ERESTARTSYS;

        if (down_interruptible(&buffer->sem))
            return -ERESTARTSYS;
    }

    /* Find and return the data. Bear in mind that the pointers could have wrapped. */
    if (buffer->writerReleased && (buffer->pWrite == buffer->pRead))
        count = 0;
    else if (!buffer->wrapAllowed || (buffer->pWrite > buffer->pRead))
        count = min(count, (size_t)(buffer->pWrite - buffer->pRead));
    else
        count = min(count, (size_t)(buffer->pEnd - buffer->pRead));

    if ( copy_to_user(buf, buffer->pRead, count) )
    {
        printk(KERN_WARNING "DspPipe_read: Unable to copy_to_user all data during read (buf %p, read %p, count %d).\n",
               buf, buffer->pRead, count);
        up(&buffer->sem);
        return -EFAULT;
    }

    buffer->pRead += count;
    if (buffer->pRead == buffer->pEnd)
        buffer->pRead = buffer->pStart;

    up(&buffer->sem);

    wake_up_interruptible(&buffer->writerQ);

    return count;
}


/* Determine how much space is left for writing */
static size_t spacefree(struct phStbDspPipe_Buffer_t *buffer)
{
    if (buffer->wrapAllowed)
        return ((buffer->pRead + buffer->bufferSize - buffer->pWrite) % buffer->bufferSize) - 1;
    else
        return (buffer->pEnd - buffer->pWrite - 1); /* -1 to ensure the pointers aren't equal */
}


/* Write data to the file */
static ssize_t phStbDspPipe_write(struct file *file, const char __user *buf, size_t count, loff_t *pos)
{
    struct phStbDspPipe_Buffer_t *buffer = (struct phStbDspPipe_Buffer_t*)file->private_data;

    DEBUG((KERN_INFO "phStbDspPipe_write: file %p, buf %p, count %d, start %p, read %p, write %p, end %p\n",
           file, buf, count, buffer->pStart, buffer->pRead, buffer->pWrite, buffer->pEnd));

    if (down_interruptible(&buffer->sem))
        return -ERESTARTSYS;

    /* Handle the situation where there is no space to write - block or return -EAGAIN */
    while (spacefree(buffer) == 0)
    {
        up(&buffer->sem);

        if (file->f_flags & O_NONBLOCK)
            return -EAGAIN;

        if (wait_event_interruptible(buffer->writerQ, (spacefree(buffer) != 0)))
            return -ERESTARTSYS;

        if (down_interruptible(&buffer->sem))
            return -ERESTARTSYS;
    }

    /* Find and return the data. Bear in mind that the pointers could have wrapped. */
    count = min(count, spacefree(buffer));
    if (buffer->pRead <= buffer->pWrite)
        count = min(count, (size_t)(buffer->pEnd - buffer->pWrite));
    else
        count = min(count, (size_t)(buffer->pRead - buffer->pWrite - 1));

    DEBUG((KERN_INFO "DspPipe_write: writing in %d bytes to %p from %p\n", count, buffer->pWrite, buf));

    if ( copy_from_user(buffer->pWrite, buf, count) )
    {
        printk(KERN_WARNING "DspPipe_write: Unable to copy_from_user all data during write.\n");
        up(&buffer->sem);
        return -EFAULT;
    }

    buffer->pWrite += count;
    if (buffer->pWrite == buffer->pEnd)
        buffer->pWrite = buffer->pStart;

    up(&buffer->sem);

    wake_up_interruptible(&buffer->readerQ);

    return count;
}


/* Support poll, select and epoll functions */
static unsigned int phStbDspPipe_poll(struct file *file, poll_table *wait)
{
    struct phStbDspPipe_Buffer_t *buffer = (struct phStbDspPipe_Buffer_t*)file->private_data;
    unsigned int mask = 0;

    down(&buffer->sem);

    /* Add both queues to the poll table. */
    poll_wait(file, &buffer->readerQ, wait);
    poll_wait(file, &buffer->writerQ, wait);

    /* If we have no writers and no data, we signal EOF. */
    if (buffer->writerReleased && (buffer->pWrite == buffer->pRead))
        mask = POLLHUP;
    else
    {
        /* Determine whether writing is possible. */
        if (spacefree(buffer) != 0)
            mask |= POLLOUT | POLLWRNORM;
        /* Determine whether reading is possible */
        if (buffer->pRead != buffer->pWrite)
            mask |= POLLIN | POLLRDNORM;
    }

    up(&buffer->sem);

    return mask;
}

/* Handle the PHSTBDSPPIPE_IOCTL_GET_WRAP_MODE ioctl */
static int phStbDspPipe_GetWrapMode(struct phStbDspPipe_Buffer_t *buffer, unsigned long arg)
{
    int *data = (int*)arg;
    DEBUG((KERN_INFO "DspPipe_GetWrapMode: buffer %p, arg %lu\n", buffer, arg));

    if (down_interruptible(&buffer->sem))
        return -ERESTARTSYS;

    *data = buffer->wrapAllowed;

    up(&buffer->sem);

    return 0;
}


/* Handle the PHSTBGPIO_IOCTL_SET_WRAP_MODE ioctl */
static int phStbDspPipe_SetWrapMode(struct phStbDspPipe_Buffer_t *buffer, unsigned long arg)
{
    DEBUG((KERN_INFO "DspPipe_SetWrapMode: buffer %p, arg %lu\n", buffer, arg));

    if (down_interruptible(&buffer->sem))
        return -ERESTARTSYS;

    buffer->wrapAllowed = (int)arg;

    up(&buffer->sem);

    return 0;
}


/* Handle all ioctls supported by the device */
static int phStbDspPipe_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    struct phStbDspPipe_Buffer_t *buffer = (struct phStbDspPipe_Buffer_t*)file->private_data;
    int result = -ENOTTY;

    switch (cmd)
    {
    case PHSTBDSPPIPE_IOCTL_SET_WRAP_MODE:
        result = phStbDspPipe_SetWrapMode(buffer, arg);
        break;

    case PHSTBDSPPIPE_IOCTL_GET_WRAP_MODE:
        result = phStbDspPipe_GetWrapMode(buffer, arg);
        break;

    default:
        break;
    }

    return result;
}


/* Handle proc reads */
static int phStbDspPipe_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
    int len = 0;

    if (offset == 0)
    {
        struct phStbDspPipe_Buffer_t *buffer = (struct phStbDspPipe_Buffer_t*)data;

        len = sprintf(buf, "start %p, read %p, write %p, end %p, size %d, wrap %d\n",
                      buffer->pStart, buffer->pRead, buffer->pWrite, buffer->pEnd, buffer->bufferSize, buffer->wrapAllowed);

        *eof = 1;
    }

    return len;
}


/* Create a character device for the specified index, add a /sys and /proc entry */
static int phStbDspPipe_CreateChrDev(unsigned int index)
{
    int result;
    dev_t devno = MKDEV(gInstance.major, index); /* Assumes that base is 0 */
    char devname[20];

    cdev_init(&gInstance.cdev[index], &phStbDspPipe_fops);
    gInstance.cdev[index].owner = THIS_MODULE;

    result = cdev_add(&gInstance.cdev[index], devno, 1);
    if (result)
    {
        printk(KERN_INFO "%s: Failed to add character device (0x%X)\n", THIS_MODULE->name, result);
        return result;
    }

    sprintf(devname, "%s%d", PHSTBDSPPIPE_DEV_NAME, index);
    class_device_create(gInstance.sysClass, NULL, devno, NULL, devname);

    sprintf(devname, "%d", index);
    create_proc_read_entry(devname, 0, gInstance.procDir, phStbDspPipe_read_proc, &gInstance.buffers[index]);

    return 0;
}


/* Delete a character device for the specified index, remove corresponding /sys and /proc entries */
static void phStbDspPipe_DelChrDev(unsigned int index)
{
    dev_t devno = MKDEV(gInstance.major, index);
    char devname[20];

    cdev_del(&gInstance.cdev[index]);

    class_device_destroy(gInstance.sysClass, devno);

    sprintf(devname, "%d", index);
    remove_proc_entry(devname, gInstance.procDir);
}


/* Initialise the module - create all devices */
static int __init phStbDspPipe_init(void)
{
    struct proc_dir_entry *baseDir;
    int result;
    int i;

    printk(KERN_INFO "%s (%s-%s)\n", THIS_MODULE_DESCRIPTION, __DATE__, __TIME__);

    if (nr_devs < 1)
        return -EINVAL;

    /* Allocate a region of minor numbers under a major number for our device */
    result = alloc_chrdev_region(&gInstance.devRegion, 0, nr_devs, PHSTBDSPPIPE_DEV_NAME);
    if (result < 0)
    {
        printk(KERN_WARNING "%s: alloc_chrdev_region failed %d\n", THIS_MODULE->name, result);
        return result;
    }

    gInstance.major = MAJOR(gInstance.devRegion);

    /* Set up /sys and /proc entries */
    gInstance.sysClass = class_create(THIS_MODULE, PHSTBDSPPIPE_DEV_NAME);
    if (gInstance.sysClass == NULL)
    {
        printk(KERN_WARNING "%s: class_create failed %d\n", THIS_MODULE->name, result);
        return -ENOMEM;
    }

    result = phStbProc_GetPNX8550ProcEntry(&baseDir);
    if (result < 0)
    {
        printk(KERN_WARNING "%s: phStbProc_GetPNX8550ProcEntry failed %d\n", THIS_MODULE->name, result);
        return result;
    }

    gInstance.procDir = proc_mkdir(PHSTBDSPPIPE_DEV_NAME, baseDir);
    if (gInstance.procDir == NULL)
    {
        printk(KERN_WARNING "%s: proc_mkdir failed\n", THIS_MODULE->name);
        return -ENOMEM;
    }

    gInstance.buffers = kmalloc(sizeof(struct phStbDspPipe_Buffer_t) * nr_devs, GFP_KERNEL);
    gInstance.cdev = kmalloc(sizeof(struct cdev) * nr_devs, GFP_KERNEL);
    if ((gInstance.buffers == NULL) || (gInstance.cdev == NULL))
    {
        kfree(gInstance.buffers);
        kfree(gInstance.cdev);
        printk(KERN_WARNING "%s: kmalloc failed during initialisation\n", THIS_MODULE->name);
    }

    memset(gInstance.buffers, 0, sizeof(struct phStbDspPipe_Buffer_t) * nr_devs);
    memset(gInstance.cdev, 0, sizeof(struct cdev) * nr_devs);

    /* Create the character devices and buffer descriptors */
    for (i = 0; (i < nr_devs) && (result == 0); i++)
    {
        struct phStbDspPipe_Buffer_t *buffer = &gInstance.buffers[i];

        result = phStbDspPipe_CreateChrDev(i);

        init_MUTEX(&buffer->sem);
        init_waitqueue_head(&buffer->readerQ);
        init_waitqueue_head(&buffer->writerQ);
    }

    return result;
}


/* Deinitialise the module - destroy all devices */
static void __exit phStbDspPipe_exit(void)
{
    struct proc_dir_entry *baseDir;
    int i;

    printk(KERN_INFO "Shutting down %s\n", THIS_MODULE->name);

    for (i = 0; i < nr_devs; i++)
    {
        phStbDspPipe_DelChrDev(i);
    }

    class_destroy(gInstance.sysClass);

    phStbProc_GetPNX8550ProcEntry(&baseDir);
    remove_proc_entry(PHSTBDSPPIPE_DEV_NAME, baseDir);

    kfree(gInstance.buffers);
    kfree(gInstance.cdev);

    unregister_chrdev_region(gInstance.devRegion, nr_devs);
}


module_init(phStbDspPipe_init);
module_exit(phStbDspPipe_exit);

