/*  Copyright (C) 2006 NXP N.V.
    All Rights Reserved.
*//**
    @file   phStbProc_BaseProcSystem.c
    @brief  Base Proc Filesystem for modules to use.

@b  Component:  phStbProc.

    Linux Kernel module that creates the following proc system

    proc/
        nxp/
            drivers/
                pnx8550/
                    video/
                    audio/
                    media/
                    i2c/.

    This allows other modules to use this module to make sure that these entries
    exist.
*//*

Rev Date        Author      Comments
--------------------------------------------------------------------------------
001 15042005    laird       First Revision
002 10052005    laird       XCheck Rework.
003 24042005    batelaan    Correct syntax error.
004 20050531    batelaan    Updated copyright.
005 20061103    laird       Moving to NXP
--------------------------------------------------------------------------------
    For consistency and standardisation retain the Section Separators.
*/

/* needed for versioned kernels (i.e. most kernels in current use) */
#include <linux/module.h>
#include <linux/kernel.h>       /* printk() */
#include <linux/proc_fs.h>
#include <phStbProc_BaseProcSystem.h>

/***********************************************
* LOCAL MACROS                                 *
* recommendation only: <MODULE><_WORD>+        *
************************************************/
#define THIS_MODULE_DESCRIPTION "NXP /proc entries driver"

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION(THIS_MODULE_DESCRIPTION);
MODULE_AUTHOR("Daniel Laird");

/******************************************************************
* STATIC FUNCTION PROTOTYPES                  <Module>_<Word>+    *
*******************************************************************/
static int __init phStbProc_ModuleInit(void);
static void __exit phStbProc_ModuleExit(void);

/******************************************************************
* LOCAL TYPEDEFS                                                  *
* recommendation only: [p]ph[<lnx|tm|NONE>]StbTemplate_<Word>+_t  *
*******************************************************************/

/*******************************************************************
* EXPORTED DATA      g[k|p|kp|pk|kpk]ph[<lnx|tm|NONE>]StbTemplate_<Word>+ *
********************************************************************/
/** Need to export the API of VideoRenderer */
EXPORT_SYMBOL(phStbProc_GetMediaProcEntry);
EXPORT_SYMBOL(phStbProc_GetVideoProcEntry);
EXPORT_SYMBOL(phStbProc_GetAudioProcEntry);
EXPORT_SYMBOL(phStbProc_GetI2cProcEntry);
EXPORT_SYMBOL(phStbProc_GetPhilipsProcEntry);
EXPORT_SYMBOL(phStbProc_GetDriversProcEntry);
EXPORT_SYMBOL(phStbProc_GetPNX8550ProcEntry);

/** Need to export the API of VideoRenderer */
module_init(phStbProc_ModuleInit);
module_exit(phStbProc_ModuleExit);

/*******************************************************************
* LOCAL DATA                                                       *
********************************************************************/
static struct proc_dir_entry *philips_base_dir;
static struct proc_dir_entry *pnx8550_base_dir;
static struct proc_dir_entry *video_base_dir;
static struct proc_dir_entry *media_base_dir;
static struct proc_dir_entry *i2c_base_dir;
static struct proc_dir_entry *audio_base_dir;
static struct proc_dir_entry *drivers_base_dir;

/*******************************************************************************
* FUNCTION IMPLEMENTATION  <Module>[_<Word>+] for static functions             *
*                          tm[<layer>]<Module>[_<Word>+] for exported functions*
********************************************************************************/

/******************************************************************************/
int phStbProc_GetDriversProcEntry(struct proc_dir_entry **procEntry)
{
    if(drivers_base_dir != NULL)
    {
        *procEntry = drivers_base_dir;
        return 0;
    }
    return -ENODATA;
}

/******************************************************************************/
int phStbProc_GetPhilipsProcEntry(struct proc_dir_entry **procEntry)
{
    if(philips_base_dir != NULL)
    {
        *procEntry = philips_base_dir;
        return 0;
    }
    return -ENODATA;
}

/******************************************************************************/
int phStbProc_GetPNX8550ProcEntry(struct proc_dir_entry **procEntry)
{
    if(pnx8550_base_dir != NULL)
    {
        *procEntry = pnx8550_base_dir;
        return 0;
    }
    return -ENODATA;
}

/******************************************************************************/
int phStbProc_GetI2cProcEntry(struct proc_dir_entry **procEntry)
{
    if(i2c_base_dir != NULL)
    {
        *procEntry = i2c_base_dir;
        return 0;
    }
    return -ENODATA;
}

/******************************************************************************/
int phStbProc_GetAudioProcEntry(struct proc_dir_entry **procEntry)
{
    if(audio_base_dir != NULL)
    {
        *procEntry = audio_base_dir;
        return 0;
    }
    return -ENODATA;
}

/******************************************************************************/
int phStbProc_GetVideoProcEntry(struct proc_dir_entry **procEntry)
{
    if(video_base_dir != NULL)
    {
        *procEntry = video_base_dir;
        return 0;
    }
    return -ENODATA;
}

/******************************************************************************/
int phStbProc_GetMediaProcEntry(struct proc_dir_entry **procEntry)
{
    if(media_base_dir != NULL)
    {
        *procEntry = media_base_dir;
        return 0;
    }
    return -ENODATA;
}

/******************************************************************************/
/** Module Creation function.  It creates all of the proc entries as specified:

    @retval int 0 if module initialised successfully.
*/
static int __init phStbProc_ModuleInit(void)
{
    philips_base_dir = proc_mkdir("nxp", NULL);
    if(philips_base_dir == NULL) {
        goto ERROR_RETURN;
    }
    drivers_base_dir = proc_mkdir("drivers", philips_base_dir);
    if(drivers_base_dir == NULL) {
        goto ERROR_RETURN;
    }
    pnx8550_base_dir = proc_mkdir("pnx8550", drivers_base_dir);
    if(pnx8550_base_dir == NULL) {
        goto ERROR_RETURN;
    }
    media_base_dir = proc_mkdir("media", pnx8550_base_dir);
    if(media_base_dir == NULL) {
        goto ERROR_RETURN;
    }
    audio_base_dir = proc_mkdir("audio", pnx8550_base_dir);
    if(audio_base_dir == NULL) {
        goto ERROR_RETURN;
    }
    video_base_dir = proc_mkdir("video", pnx8550_base_dir);
    if(video_base_dir == NULL) {
        goto ERROR_RETURN;
    }
    i2c_base_dir = proc_mkdir("i2c", pnx8550_base_dir);
    if(i2c_base_dir == NULL) {
        goto ERROR_RETURN;
    }
    printk ("%s (%s-%s)\n", THIS_MODULE_DESCRIPTION, __DATE__, __TIME__);
    return 0;
  ERROR_RETURN:
    return -ENOMEM;
}


/******************************************************************************/
/** Remove the proc entries and de-init the module
*/
static void __exit phStbProc_ModuleExit(void)
{
    remove_proc_entry("i2c", pnx8550_base_dir);
    remove_proc_entry("video", pnx8550_base_dir);
    remove_proc_entry("audio", pnx8550_base_dir);
    remove_proc_entry("media", pnx8550_base_dir);

    remove_proc_entry("pnx8550", drivers_base_dir);
    remove_proc_entry("drivers", philips_base_dir);
    remove_proc_entry("nxp", NULL);
    printk("Shutting down %s\n", THIS_MODULE_DESCRIPTION);
}
