//@+leo-ver=4
//@+node:@file host/procfs.c
//@@language c
/**
 @file
 */

/**
 * @defgroup group_linux Luddite Linux Side Library
 * These routines are made available by lud-core to all luddite drivers
 * and linux side modules.
 * @{
 */

/* Copyright Philips Research Labs, Redhill, UK. 2001
 *
 * Prototype Software
 *
 * Contact: Tim Everett (tim.everett@philips.com)
 *
 */

/***** Includes *****/

//@<< lud-base #includes >>
//@+node:<< lud-base #includes >>
#ifdef CONFIG_SMP
#define __SMP__
#endif

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/page.h>
#include <asm/semaphore.h>
#include <asm/atomic.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>

#include <linux/module.h>

#include <linux/version.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/kmod.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#include <linux/device.h>
#include <linux/cdev.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18))
#include <linux/compiler.h>
#else
#define likely(x) (x)
#define unlikely(x) (x)
#endif /* LINUX_VERSION_CODE */

#include <stdarg.h>


#include "compat.h"
#include "lud_mmio.h"
#include "luddite-host.h"
#include "luddite/version.h"
#include "luddite/os.h"
#include "exported.h"

#include "core.h"
#include "procfs.h"
#include "node.h"
#include "helpers.h"
#include "channels.h"
//@-node:<< lud-base #includes >>
//@nl
//@<< lud-proc declarations >>
//@+node:<< lud-proc declarations >>
static spinlock_t dbg_sem;
static int lud_dbg[MAX_LUD_DBG][4];
static int dbg_w = 0;
//@-node:<< lud-proc declarations >>
//@nl
//@+others
//@+node:dbg_add
//@+at
//@nonl
// * * This function adds a row of data to the cyclic buffer viewable via the
// /proc * interface. * @param a first column data item * @param b first
// column data item * @param c first column data item * @param d first column
// data item
//@-at
//@@c

void dbg_add(unsigned int a,unsigned int b,unsigned int c,unsigned int d)
{
	unsigned long irqstate;
	spin_lock_irqsave(&dbg_sem,irqstate);
	lud_dbg[dbg_w][0] = a;
	lud_dbg[dbg_w][1] = b;
	lud_dbg[dbg_w][2] = c;
	lud_dbg[dbg_w][3] = d;
	dbg_w = (dbg_w+1) % MAX_LUD_DBG;

	spin_unlock_irqrestore(&dbg_sem,irqstate);
}
//@-node:dbg_add
//@+node:lud_read_procmod

static struct proc_dir_entry* lud_proc_root = NULL;
static struct semaphore proc_sem;

/* ************************************************************************ *
 * /proc query functions                                                    *
 * ************************************************************************ */

static int lud_read_procmod(char* page,char **start, off_t offset,
				int count,int* eof,void* data)
{
	int i,len = 0;
	struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;

	if (down_interruptible(&tmd->mod_sem)) {
		return -1;
	}

	for (i=0;i<LUDMOD_MAX_MODULES;i++) {
		if (tmd->lud_modules[i].name) {
			len += sprintf(page+len,"%-20s %-40s %d\n",
				tmd->lud_modules[i].name,
				tmd->lud_modules[i].opts,
				tmd->lud_modules[i].pri);
		}
	}
	*eof = 1;
	up(&tmd->mod_sem);
	return len;
}
//@-node:lud_read_procmod
//@+node:lud_read_procdbg

static int lud_read_procdbg(char* page,char **start, off_t offset,
				int count,int* eof,void* data)
{
	int i;
	int len = 0;
	for (i=0;i<64;i++) {
		len += sprintf(page+len,"0x%08x 0x%08x 0x%08x 0x%08x\n",
			lud_dbg[(i+dbg_w)%64][0],
			lud_dbg[(i+dbg_w)%64][1],
			lud_dbg[(i+dbg_w)%64][2],
			lud_dbg[(i+dbg_w)%64][3]);
	}
	*eof = 1;
	return len;
}
//@-node:lud_read_procdbg
//@+node:lud_read_procstatus

static int lud_read_procstatus(char* page,char **start, off_t offset,
				int count,int* eof,void* data)
{
	int i;
	int len = 0;
	struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;
	unsigned int* status = (unsigned int*)tmd->luddev->status;
	/* status is one cache row ... 16 x 32 bit ints */

	/* May not have initialized the Target yet, so status may be NULL. */
	if (status != NULL) {
		for (i=0;i<16;i++) {
			len += sprintf(page+len,"%2d 0x%08x\n",i,readl((IOTYPE)&status[i]));
		}
	}
	*eof = 1;
	return len;
}
//@-node:lud_read_procstatus
//@+node:lud_read_proctm

static int lud_read_proctm(char* page,char **start, off_t offset,
				int count,int* eof,void* data)
{
	int len = 0;
	struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;

	const char * state_string;

	switch (tmd->luddev->state) {
		case LUD_UNDEFINED:
			state_string = "Undefined"; break;
		case LUD_RESET:
			state_string = "Reset"; break;
		case LUD_RUNNING:
			state_string = "Running something (not Luddite?)"; break;
		case LUD_LUDDITE_STARTING:
			state_string = "Luddite starting"; break;
		case LUD_LUDDITE:
			state_string = "Running Luddite"; break;
		case LUD_CRASHED:
			state_string = "Crashed"; break;
		default:
			state_string = "Unknown"; break;
	}

	len += sprintf(page+len,
		"Kind:   %s (lud_type = %d)\n"
		"Base:   0x%p\n"
		"Memory: 0x%08x bytes (%d MB)\n"
		"Freq:   %d Hz\n"
		"State:  %s\n"
		"Printk: %d\n",
		tmd->luddev->name, tmd->luddev->lud_type, (void*)tmd->luddev->sdram_phy,
		tmd->luddev->sdram_size, tmd->luddev->sdram_size >> 20,
		tmd->luddev->freq,
		state_string,
		tmd->dprintk_level);


	*eof = 1;
	return len;
}
//@-node:lud_read_proctm
//@+node:lud_read_procpools

static int lud_read_procpools(char* page,char **start, off_t offset,
				int count,int* eof,void* data)
{
	int len = 0;
	int blen = 0;
	char *buf;
	struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;
	if(down_interruptible(&proc_sem))
		return -ERESTARTSYS;
	if (tmd->luddev->state == LUD_LUDDITE){
		LUD_DEV_LINK * pool_link = lud_clone_link(tmd->loader);
		unsigned int addr = lud_send_request(pool_link,LUD_SYS_POOLS,0,0,0,0);
		buf = (char*)lud_remap_addr(tmd->luddev,addr);
		if(buf == NULL){
			len +=sprintf(page+len,"Memory mapping failed\n");
		}else{
			char *bounce_buf = kmalloc(PROC_BUFFER_SIZE, GFP_KERNEL);
			if (unlikely(!bounce_buf)) {
				return -ENOMEM;
			}
			blen = strncpy_fromio(bounce_buf,buf,PROC_BUFFER_SIZE);
			if(blen >4000){
				len += sprintf(page+len,"We might have a problem with the buffer length\n");
			} else {
				len += sprintf(page+len,"%s",bounce_buf);
			}
			kfree(bounce_buf);
		}
		lud_free_link(pool_link);
	}
	up(&proc_sem);
	*eof = 1;
	return len;
}
//@-node:lud_read_procpools
//@+node:lud_read_proctasks

static int lud_read_proctasks(char* page,char **start, off_t offset,
				int count,int* eof,void* data)
{
	int len = 0;
	int blen = 0;
	char *buf;
	struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;
	if(down_interruptible(&proc_sem))
		return -ERESTARTSYS;
	if (tmd->luddev->state == LUD_LUDDITE){
		LUD_DEV_LINK * task_link = lud_clone_link(tmd->loader);
		unsigned int addr = lud_send_request(task_link,LUD_SYS_TASKS,0,0,0,0);
		buf = (char*)lud_remap_addr(tmd->luddev,addr);
		if(buf == NULL){
			len +=sprintf(page+len,"Memory mapping failed\n");
		}else{
			char *bounce_buf = kmalloc(PROC_BUFFER_SIZE, GFP_KERNEL);
			if (unlikely(!bounce_buf)) {
				return -ENOMEM;
			}
			blen = strncpy_fromio(bounce_buf,buf,PROC_BUFFER_SIZE);
			if(blen >4000){
				len += sprintf(page+len,"We might have a problem with the buffer length\n");
			} else {
				len += sprintf(page+len,"%s",bounce_buf);
			}
			kfree(bounce_buf);
		}
		lud_free_link(task_link);
	}
    up(&proc_sem);
	*eof = 1;
	return len;
}
//@-node:lud_read_proctasks
//@+node:lud_read_procqueues

static int lud_read_procqueues(char* page,char **start, off_t offset,
				int count,int* eof,void* data)
{
	int len = 0;
	int blen = 0;
	char *buf;
	struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;
	if(down_interruptible(&proc_sem))
		return -ERESTARTSYS;
	if (tmd->luddev->state == LUD_LUDDITE){
		LUD_DEV_LINK * queue_link = lud_clone_link(tmd->loader);
		unsigned int addr = lud_send_request(queue_link,LUD_SYS_QUEUES,0,0,0,0);
		buf = (unsigned char*)lud_remap_addr(tmd->luddev,addr);
		if(buf == NULL){
			len +=sprintf(page+len,"Memory mapping failed\n");
		}else{
			char *bounce_buf = kmalloc(PROC_BUFFER_SIZE, GFP_KERNEL);
			if (unlikely(!bounce_buf)) {
				return -ENOMEM;
			}
			blen = strncpy_fromio(bounce_buf,buf,PROC_BUFFER_SIZE);
			len += sprintf(page+len,"%s",bounce_buf);
			kfree(bounce_buf);
		}
		lud_free_link(queue_link);
	}
	up(&proc_sem);
	*eof = 1;
	return len;
}
//@-node:lud_read_procqueues
//@+node:lud_read_procisrs

static int lud_read_procisrs(char* page,char **start, off_t offset,
				int count,int* eof,void* data)
{
	int len = 0;
	int blen = 0;
	char *buf;
	struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;
	if(down_interruptible(&proc_sem))
		return -ERESTARTSYS;
	if (tmd->luddev->state == LUD_LUDDITE){
		LUD_DEV_LINK * queue_link = lud_clone_link(tmd->loader);
		unsigned int addr = lud_send_request(queue_link,LUD_SYS_ISRS,0,0,0,0);
		buf = (unsigned char*)lud_remap_addr(tmd->luddev,addr);
		if(buf == NULL){
			len +=sprintf(page+len,"Memory mapping failed\n");
		}else{
			char *bounce_buf = kmalloc(PROC_BUFFER_SIZE, GFP_KERNEL);
			if (unlikely(!bounce_buf)) {
				return -ENOMEM;
			}
			blen = strncpy_fromio(bounce_buf,buf,PROC_BUFFER_SIZE);
			len += sprintf(page+len,"%s",bounce_buf);
			kfree(bounce_buf);
		}
		lud_free_link(queue_link);
	}
	up(&proc_sem);
	*eof = 1;
	return len;
}
//@-node:lud_read_procisrs

static int lud_read_procrt(char* page,char **start, off_t offset,
                                int count,int* eof,void* data)
{
	unsigned int mod = 0;
	unsigned int t_1 = 0;
	unsigned int t_2 = 0;
	unsigned int t_3 = 0;
	int len = 0;
	struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;
	if(down_interruptible(&proc_sem))
		return -ERESTARTSYS;
	if (tmd->luddev->state == LUD_LUDDITE){
		unsigned int forward = 0;
		unsigned int backward = 0;
		LUD_DEV_LINK* link = lud_clone_link(tmd->loader);
		/* Find a timer running at the clock frequency */
		int found = 0;
		unsigned int tbase = TIMER1_TMODULUS;

		for (found=0;found<4;found++,tbase += 0x20) {
			unsigned int c = MMIO_TMread(tmd->luddev,tbase+0x8);
			if (((c&0xf01)==1)|| /* running and source == clock */
			    ((c&0xf0f01)==0x101)) /* source == pclock, p=1 */
				break;
		}
		if (found<4) {
			t_1 = MMIO_TMread(tmd->luddev,tbase+0x4);
			mb();
			t_2 = lud_send_request(link,LUD_SYS_ROUNDTRIP,found,0,0,0);
			mb();
			t_3 = MMIO_TMread(tmd->luddev,tbase+0x4);
			mod = MMIO_TMread(tmd->luddev,tbase);

			if (t_2 < t_1) {
				forward = (mod + t_2) - t_1;
			} else {
				forward = t_2 - t_1;
			}

			if (t_3 < t_2) {
				backward = (mod + t_3) - t_2;
			} else {
				backward = t_3 - t_2;
			}

			len += sprintf(page+len,"%12u %12u %12u\n",forward,backward,tmd->luddev->freq);
		} else {
			len += sprintf(page+len,"No system clock found\n");
		}

		lud_free_link(link);
	}
	up(&proc_sem);
	*eof = 1;
	return len;
}

//@+node:lud_read_procmemory



static int lud_read_procmemory(char* page,char **start, off_t offset,
				int count,int* eof,void* data)
{
	int len = 0;
	int blen = 0;
	char *buf;
	struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;
	if(down_interruptible(&proc_sem))
		return -ERESTARTSYS;
	if (tmd->luddev->state == LUD_LUDDITE){
		LUD_DEV_LINK * memory_link = lud_clone_link(tmd->loader);
		unsigned int addr = lud_send_request(memory_link,LUD_SYS_MEMORY,0,0,0,0);
		buf = (unsigned char*)lud_remap_addr(tmd->luddev,addr);
		if(buf == NULL){
			len +=sprintf(page+len,"Memory mapping failed\n");
		}else{
			char *bounce_buf = kmalloc(PROC_BUFFER_SIZE, GFP_KERNEL);
			if (unlikely(!bounce_buf)) {
				return -ENOMEM;
			}
			blen = strncpy_fromio(bounce_buf,buf,PROC_BUFFER_SIZE);
			len += sprintf(page+len,"%s",bounce_buf);
			kfree(bounce_buf);
		}
		lud_free_link(memory_link);
	}
	up(&proc_sem);
	*eof = 1;
	return len;
}
//@-node:lud_read_procmemory
//@+node:lud_write_procmemory

static int lud_write_procmemory(struct file *filp,
                               const char *buffer,
                               unsigned long count,
                               void *data)
{
	/* Only do this on first write to file */
	if (filp->f_pos == 0) {
		struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;
		if(down_interruptible(&proc_sem))
			return -ERESTARTSYS;
		if (tmd->luddev->state == LUD_LUDDITE){
			LUD_DEV_LINK * memory_link = lud_clone_link(tmd->loader);
			lud_send_request(memory_link,LUD_SYS_MEMORY,1,0,0,0);
			lud_free_link(memory_link);
		}
		up(&proc_sem);
	}

	if (count > 0x7fffffffu) {
		count = 0x7fffffffu; /* clip to int */
	}
	filp->f_pos += count;

	return count;
}

static int lud_read_procevents(char* page,char **start, off_t offset,
				int count,int* eof,void* data)
{
	int len = 0;
	struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;
	if(down_interruptible(&proc_sem))
		return -ERESTARTSYS;
	if (tmd->luddev->state == LUD_LUDDITE){
		LUD_DEV_LINK * memory_link = lud_clone_link(tmd->loader);
		unsigned int val = lud_send_request(memory_link,LUD_SYS_GET_DEBUG_EVENT,0,0,0,0);
		len += sprintf(page+len,"%d\n",val);
		lud_free_link(memory_link);
	}
	up(&proc_sem);
	*eof = 1;
	return len;
}

static int lud_write_procevents(struct file *filp,
                               const char *buffer,
                               unsigned long count,
                               void *data)
{
	/* Only do this on first write to file */
	if (filp->f_pos == 0) {
		struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;
		if(down_interruptible(&proc_sem))
			return -ERESTARTSYS;
		if (tmd->luddev->state == LUD_LUDDITE){
			unsigned int val=1;
			LUD_DEV_LINK * memory_link = lud_clone_link(tmd->loader);
			sscanf(buffer,"%d",&val);
			lud_send_command(memory_link,LUD_SYS_SET_DEBUG_EVENT,val,0,0,0);
			lud_free_link(memory_link);
		}
		up(&proc_sem);
	}

	if (count > 0x7fffffffu) {
		count = 0x7fffffffu; /* clip to int */
	}
	filp->f_pos += count;

	return count;
}
//@-node:lud_write_procmemory
//@+node:lud_write_procdebug

#define MEMEV 14

static int lud_write_procdebug(struct file *filp,
				const char *buffer,
				unsigned long count,
				void *data)
{
	if (filp->f_pos == 0) {
		struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;
		int inclu=0;
		unsigned int ev1 = 0;
		unsigned int ev2 = 0;
		unsigned int ev3 = 0;
		int	     trace=0;
		switch (sscanf(buffer,"%d%x%x%d",&ev1,&ev2,&ev3,&trace)) {
		case 1:
			/* activate oob detect */
			MMIO_TMwrite(tmd->luddev->sdram_phy,tmd->luddev,BDATAALOW);
			MMIO_TMwrite(tmd->luddev->mmio_phy+2*(1024*1024),tmd->luddev,BDATAAHIGH);

			if (ev1>=1 && ev1<=3) {
				MMIO_TMwrite(0x0,tmd->luddev,TIMER4_CTL);
				mb();
				MMIO_TMwrite(0,tmd->luddev,TIMER4_TVALUE);
				mb();
				MMIO_TMwrite(0x101 | (ev1<<2),tmd->luddev,BDCTL);
				mb();
				MMIO_TMwrite(0,tmd->luddev,TIMER4_TMODULUS);
				mb();
				MMIO_TMwrite(0x10001 | (3<<8),tmd->luddev,TIMER4_CTL);
				mb();
			} else {
				MMIO_TMwrite(0,tmd->luddev,BDCTL);
				mb();
				MMIO_TMwrite(0,tmd->luddev,TIMER4_CTL);
				mb();
			}
			break;
		case 2:
			if ((ev1>=0) && (ev1<MEMEV) && (ev2>=0) && (ev2 <= MEMEV)) {
				MMIO_TMwrite((ev1 | (ev2<<4)),tmd->luddev,MEM_EVENTS);
				mb();
				MMIO_TMwrite(0x10001 | (5<<8),tmd->luddev,TIMER2_CTL);
				mb();
				MMIO_TMwrite(0x10001 | (6<<8),tmd->luddev,TIMER3_CTL);
				mb();

			}
			break;
		case 3:
			trace = 0;
			/* fall through */
		case 4:
			/* activate oob detect */

			if (ev1>3) {
				inclu=1;
				ev1 = ev1&0x3;
			}
			if (ev1>=1 && ev1<=3) {
				MMIO_TMwrite(ev2,tmd->luddev,BDATAALOW);
				MMIO_TMwrite(ev3,tmd->luddev,BDATAAHIGH);
				MMIO_TMwrite(0x0,tmd->luddev,TIMER4_CTL);
				mb();
				MMIO_TMwrite(0,tmd->luddev,TIMER4_TVALUE);
				mb();
				if (inclu) {
					MMIO_TMwrite(0x001 | (ev1<<2),tmd->luddev,BDCTL);
				} else {
					MMIO_TMwrite(0x101 | (ev1<<2),tmd->luddev,BDCTL);
				}
				mb();
				MMIO_TMwrite(trace,tmd->luddev,TIMER4_TMODULUS);
				mb();
				MMIO_TMwrite(0x10001 | (3<<8),tmd->luddev,TIMER4_CTL);
				mb();
			} else {
				MMIO_TMwrite(0,tmd->luddev,BDCTL);
				MMIO_TMwrite(0,tmd->luddev,TIMER4_CTL);
			}
			break;
		}


		MMIO_TMwrite(0,tmd->luddev,TIMER2_TVALUE);
		MMIO_TMwrite(0,tmd->luddev,TIMER3_TVALUE);
		MMIO_TMwrite(0,tmd->luddev,TIMER4_TVALUE);
	}
	if (count > 0x7fffffffu) {
		count = 0x7fffffffu; /* clip to int */
	}
	filp->f_pos += count;
	return count;
}
//@-node:lud_write_procdebug
//@+node:lud_read_procdebug

static char* memevents[] = {
	"None",
	"I-cache misses",
	"I-cache stalls",
	"D-cache bank conflicts",
	"D-cache read misses",
	"D-cache write misses",
	"D-cache stalls (x I)",
	"D-cache copyback",
	"Copyback full",
	"D-cache write miss FU occupied",
	"D-cache stream miss",
	"Prefetch started not discarded",
	"Prefetch discarded",
	"Prefetch discarded (hits cache)",
	"Reserved",
	"Reserved"
};

static int lud_read_procdebug(char* page,char **start, off_t offset,
				int count,int* eof,void* data)
{
	unsigned int bdctl;
	int len = 0;
	struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;
	unsigned int mem_events = MMIO_TMread(tmd->luddev,MEM_EVENTS);

	int ev1 = mem_events&0xf;
	int ev2 = (mem_events>>4)&0xf;
	unsigned int t1 = MMIO_TMread(tmd->luddev,TIMER2_TVALUE);
	unsigned int t2 = MMIO_TMread(tmd->luddev,TIMER3_TVALUE);
	unsigned int t3 = MMIO_TMread(tmd->luddev,TIMER4_TVALUE);

	MMIO_TMwrite(0,tmd->luddev,TIMER2_TVALUE);
	MMIO_TMwrite(0,tmd->luddev,TIMER3_TVALUE);
	MMIO_TMwrite(0,tmd->luddev,TIMER4_TVALUE);
	if (ev1)
		len +=sprintf(page+len,"%-40s= %u\n",memevents[ev1],t1);
	if (ev2)
		len +=sprintf(page+len,"%-40s= %u\n",memevents[ev2],t2);
	if ((bdctl=MMIO_TMread(tmd->luddev,BDCTL))) {
		char what[128];
		switch ((bdctl>>2)&3) {
			case 0: sprintf(what,"none"); break;
			case 1: sprintf(what,"reads"); break;
			case 2: sprintf(what,"writes"); break;
			case 3: sprintf(what,"reads/writes"); break;
		}
		if (MMIO_TMread(tmd->luddev,TIMER4_TMODULUS)) {
			strcat(what," interrupt");
		} else {
			strcat(what," count");
		}
		if ((bdctl>>8)&1) {
			strcat(what," outside range");
		} else {
			strcat(what," inside range");
		}
		len +=sprintf(page+len,"%-40s= %u (range 0x%x to 0x%x)\n",what,t3,MMIO_TMread(tmd->luddev,BDATAALOW),MMIO_TMread(tmd->luddev,BDATAAHIGH));
	}
	*eof = 1;
	return len;
}
//@-node:lud_read_procdebug
//@+node:lud_read_procnodes

#define LUD_PROC_NODES_LINE_LENGTH (5+2+32+1+10+1+10+1+5+1)

/**
 * Read from /proc/driver/luddite/0/nodes.
 *
 * @param buf    Output buffer, kernel space, size = PAGE_SIZE
 * @param start  On exit, *start will be set to the start of the data within buf.
 * @param offset Offset into the file to read from.
 * @param count  Maximum number of bytes to read.
 * @param eof    If EOF is encountered, *eof will be set to nonzero.
 * @param data   Unused.
 * @return       Number of bytes read.
 */
static int lud_read_procnodes(char* buf,char **start, off_t offset,
				int count,int* eof,void* data)
{
	struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;
	struct LUD_NODE_DEV * node;
	int node_id;
	int len;
	int line_len;
	int lines_to_skip;
	int bytes_at_beginning_to_skip;
	char handler[11];

	lines_to_skip = offset / LUD_PROC_NODES_LINE_LENGTH;
	bytes_at_beginning_to_skip = offset % LUD_PROC_NODES_LINE_LENGTH;

	/*printk(KERN_CRIT "lud_read_procnodes(): %d bytes at offset %d (== %d lines * %d bytes/line + %d bytes)\n",
		count, (int)offset, lines_to_skip, LUD_PROC_NODES_LINE_LENGTH, bytes_at_beginning_to_skip);*/

	len = 0;

	if (lines_to_skip > 0)
	{
		lines_to_skip--;
	}
	else
	{
		line_len = sprintf(buf + len,
				"%5s: %-32s %-10s %-10s %-5s\n",
				"ID", "name", "<handler>", "flags", "xfers");
		if (line_len != LUD_PROC_NODES_LINE_LENGTH) {
			printk(KERN_EMERG "line_len(header)=%d, LUD_PROC_NODES_LINE_LENGTH=%d\n", line_len, LUD_PROC_NODES_LINE_LENGTH);
			BUG();
		}
		len += line_len;
	}

	for (node_id = 0; node_id < LUD_MAX_NODES; node_id++)
	{
		node = tmd->lud_nodes[node_id];
		if (node)
		{
			if (lines_to_skip > 0)
			{
				lines_to_skip--;
			}
			else
			{
				if (len + LUD_PROC_NODES_LINE_LENGTH + 1 > PAGE_SIZE) {
					/* File too long fit in buffer */
					goto done;
				}

				if (node->handler) {
					sprintf(handler, "<%08x>", (int)((long)node->handler));
				} else {
					strcpy(handler, " none");
				}

				line_len = sprintf(buf + len,
						"%5d: %-32.32s %-10s %#010x %-5s\n",
						node_id,
						node->name,
						handler,
						node->flags,
						(node->read
							? (node->write ? "both" : "read")
							: (node->write ? "write" : "none")
						)
					);

				if (line_len != LUD_PROC_NODES_LINE_LENGTH) {
					printk(KERN_EMERG "line_len=%d, LUD_PROC_NODES_LINE_LENGTH=%d\n", line_len, LUD_PROC_NODES_LINE_LENGTH);
					BUG();
				}
				len += line_len;
			}
		}
	}

	*eof = 1;

done:

	/*printk(KERN_CRIT "lud_read_procnodes(): generated %d of %d bytes\n", len, count);*/

	if (len <= bytes_at_beginning_to_skip)
	{
		len = 0;
	}
	else if (bytes_at_beginning_to_skip != 0)
	{
		len -= bytes_at_beginning_to_skip;
		memmove(buf, buf + bytes_at_beginning_to_skip, len);
	}

	if (len > count)
	{
		/*printk(KERN_CRIT "lud_read_procnodes(): clipping %d to %d bytes\n", len, count);*/
		len = count;
	}

	/* Tell procfs where we put the data.  Also tells procfs that this is a "clever"
	 * driver that can return partial files.  (We don't just always return the whole file
	 * and let procfs sort it out, because this file is potentially too big).
	 */
	*start = buf;

	/*printk(KERN_CRIT "lud_read_procnodes(): done %d of %d bytes\n", len, count);*/

	return len;
}
//@-node:lud_read_procnodes
//@+node:lud_read_procproperties

static int lud_read_procproperties(char* page,char **start, off_t offset,
				int count,int* eof,void* data)
{
	int len = 0;
	int blen = 0;
	char *buf;
	struct LUD_DEV_STRUCT* tmd = (struct LUD_DEV_STRUCT*)data;
	if(down_interruptible(&proc_sem))
		return -ERESTARTSYS;
	if (tmd->luddev->state == LUD_LUDDITE){
		LUD_DEV_LINK *properties_link = lud_clone_link(tmd->loader);
		unsigned int addr = lud_send_request(properties_link,LUD_SYS_PROPERTIES,0,0,0,0);
		buf = (unsigned char*)lud_remap_addr(tmd->luddev,addr);
		if(buf == NULL){
			len +=sprintf(page+len,"Memory mapping failed\n");
		}else{
			char *bounce_buf = kmalloc(PROC_BUFFER_SIZE, GFP_KERNEL);
			if (unlikely(!bounce_buf)) {
				return -ENOMEM;
			}
			blen = strncpy_fromio(bounce_buf,buf,PROC_BUFFER_SIZE);
			len += sprintf(page+len,"%s",bounce_buf);
			kfree(bounce_buf);
		}
		lud_free_link(properties_link);
	}
	up(&proc_sem);
	*eof = 1;
	return len;
}
//@-node:lud_read_procproperties
//@+node:create_proc_files

void	create_proc_files(LUD_DEV* dev)
{
	int i;
	int index = -1;
	char procname[128];

	for (i=0;i<MAX_LUD_DEV;i++) {
		if (lud_devices[i].luddev == dev) {
			index = i;
			break;
		}
	}

	if (index<0) {
		return;
	}

	if (lud_proc_root) {
		sprintf(procname,"%d",index);
		dev->proc_root = create_proc_entry(procname,S_IFDIR,lud_proc_root);
		if (dev->proc_root) {
			create_proc_read_entry("modules",0,dev->proc_root,
				lud_read_procmod,&lud_devices[index]);
			create_proc_read_entry("tm",0,dev->proc_root,
				lud_read_proctm,&lud_devices[index]);
			create_proc_read_entry("tasks",0,dev->proc_root,
				lud_read_proctasks,&lud_devices[index]);
			create_proc_read_entry("pools",0,dev->proc_root,
				lud_read_procpools,&lud_devices[index]);
			/* create_proc_read_entry("memory",0,dev->proc_root,
				lud_read_procmemory,&lud_devices[index]); */
			{
				struct proc_dir_entry *proc_memory;
				proc_memory = create_proc_entry("memory",0,dev->proc_root);
				if (proc_memory != NULL) {
					proc_memory->data = &lud_devices[index];
					proc_memory->read_proc = lud_read_procmemory;
					proc_memory->write_proc = lud_write_procmemory;
				}
			}
			create_proc_read_entry("nodes",0,dev->proc_root,
				lud_read_procnodes,&lud_devices[index]);
			create_proc_read_entry("properties",0,dev->proc_root,
				lud_read_procproperties,&lud_devices[index]);
			create_proc_read_entry("status",0,dev->proc_root,
				lud_read_procstatus,&lud_devices[index]);
			create_proc_read_entry("queues",0,dev->proc_root,
				lud_read_procqueues,&lud_devices[index]);
			create_proc_read_entry("interrupts",0,dev->proc_root,
				lud_read_procisrs,&lud_devices[index]);
			create_proc_read_entry("roundtrip",0,dev->proc_root,
				lud_read_procrt,&lud_devices[index]);
			{
				struct proc_dir_entry *proc_debug;
				proc_debug = create_proc_entry("debug",0,dev->proc_root);
				if (proc_debug != NULL) {
					proc_debug->data = &lud_devices[index];
					proc_debug->read_proc = lud_read_procdebug;
					proc_debug->write_proc = lud_write_procdebug;
				}
			}
			{
				struct proc_dir_entry *proc_events;
				proc_events = create_proc_entry("events",0,dev->proc_root);
				if (proc_events != NULL) {
					proc_events->data = &lud_devices[index];
					proc_events->read_proc = lud_read_procevents;
					proc_events->write_proc = lud_write_procevents;
				}
			}
		}
	}
}
//@-node:create_proc_files
//@+node:remove_proc_files

void remove_proc_files(LUD_DEV* dev)
{
	int i;
	int index = -1;
	char procname[128];

	for (i=0;i<MAX_LUD_DEV;i++) {
		if (lud_devices[i].luddev == dev) {
			index = i;
			break;
		}
	}

	if (index<0) {
		return;
	}

	if (dev->proc_root) {
		remove_proc_entry("modules",dev->proc_root);
		remove_proc_entry("tm",dev->proc_root);
		remove_proc_entry("tasks",dev->proc_root);
		remove_proc_entry("pools",dev->proc_root);
		remove_proc_entry("memory",dev->proc_root);
		remove_proc_entry("nodes",dev->proc_root);
		remove_proc_entry("properties",dev->proc_root);
		remove_proc_entry("status",dev->proc_root);
		remove_proc_entry("debug",dev->proc_root);
		remove_proc_entry("queues",dev->proc_root);
		remove_proc_entry("interrupts",dev->proc_root);
		remove_proc_entry("roundtrip",dev->proc_root);
		remove_proc_entry("events",dev->proc_root);
		sprintf(procname,"%d",index);
		remove_proc_entry(procname,lud_proc_root);
	}
}
//@-node:remove_proc_files
//@+node:init_proc_system

void init_proc_system(void)
{
	spin_lock_init(&dbg_sem);
	sema_init(&proc_sem,1);
	lud_proc_root = create_proc_entry("driver/luddite",S_IFDIR,0);
	create_proc_read_entry("dbg",0,lud_proc_root,
                                lud_read_procdbg,NULL);
}
//@-node:init_proc_system
//@+node:clean_proc_system

void clean_proc_system(void)
{
	remove_proc_entry("driver/luddite/dbg",0);
	remove_proc_entry("driver/luddite",0);
}
//@-node:clean_proc_system
//@-others
//@-node:@file host/procfs.c
//@-leo
