#ifndef PHSTBRPCTM_H    /* Multi-include protection.   <FILENAME>_H */
#define PHSTBRPCTM_H

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright (C) 2006 NXP B.V.
 * All Rights Reserved.
*//**
    @file   phStbRpc.h
    @brief  TM Server for RPC.

@b  Component: phStbRpc.

*//*

Rev Date        Author      Comments
--------------------------------------------------------------------------------
001 20050418    batelaan    Original
002 20050511    batelaan    Added error codes; phStbRpcTM_HandlerTaskFunc_t.
003 20050512    batelaan    Crosscheck fixes.
004 20050517    batelaan    Crosscheck fixes.
005 20070730    laird       Updated Header
--------------------------------------------------------------------------------
    For consistency and standardisation retain the Section Separators.
*/


/*******************
* INCLUDE FILES    *
********************/

#include <phStbCommon.h>
#include <phStbRpc_Common.h>
#include <tmosal.h>

/**************************************************************
* EXPORTED MACROS         PH[<LNX|TM|NONE>]_IPSTBRPC_<_WORD>+        *
***************************************************************/

/** Insufficient resource (unknown reason). */
#define PH_ERR_STBRPCTM_NO_RESOURCE 1

/** Insufficient memory. */
#define PH_ERR_STBRPCTM_NO_MEM 2

/** RPC Handler is not initialised. */
#define PH_ERR_STBRPCTM_NOT_INIT 3

/** Illegal handler Id (negative or >= #. */
#define PH_ERR_STBRPCTM_HANDLERID 4

/** No rpcEntries or unmarshal function specified in call to phStbRpcTM_HandlerCreate() error. */
#define PH_ERR_STBRPCTM_RPCDEF 5

/** An entry to initialise a ::phStbRpcTM_RpcEntry_t
    for a table of Rpc entries.
    This macro expands to:
    - rpcId
    - rpcId as a C string
    - a function name consisting of rpcId followed by _Unmarshal.
    @param rpcId An Rpc id enum value belonging to a certain Rcp handler (see ::phStbRpc_HandlerId_t).
*/
#define PH_STBRPCTM_RPCENTRY(rpcId) {rpcId, #rpcId, rpcId ## _Unmarshal}

/** Define an entry to initialise a ::phStbRpcTM_RpcEntry_t
    for a table of Rpc entries.
    This macro expands to:
    - rpcId
    - rpcId as a C string
    - pUnmarshalFunc
    @param rpcId An Rpc id enum value belonging to a certain Rcp handler (see ::phStbRpc_HandlerId_t).
    @param pUnmarshalFunc Pointer to the unmarshall function.
    @see PH_STBRPCTM_RPCENTRY
*/
#define PH_STBRPCTM_RPCENTRY2(rpcId, pUnmarshalFunc) {rpcId, #rpcId, pUnmarshalFunc}

/**************************************************************
* EXPORTED TYPEDEFS       [p]ph[<lnx|tm|NONE>]StbRpcTM_<Word>+_t  *
***************************************************************/


/** A function called by an Rpc handler on the target cpu, used
    to unmarshall the buffer, extracting arguments of the function to call
    (given by rpcId).
    The function is allowed to update the contents of the buffer,
    which are returned to the calling cpu.
    @param handlerId @b IN Rpc handler Id.
    @param rpcId @b IN Rpc function Id within this handler.
    @param pBuffer @b INOUT Buffer with data to be transferred to the target. This
    will be updated by the RPC call.
    @param pBufferSize @b INOUT Size of the payload data buffer. For an synchronous
    RPC function this needs to be updated with the actual size of data returned via
    pBuffer.
    @return any value.
*/
typedef UInt32 phStbRpcTM_RpcUnmarshalFunc_t(
    phStbRpc_HandlerId_t handlerId,
    UInt32 rpcId,
    pUInt8 pBuffer,
    pUInt32 pBufferSize
);


/** A function called by an Rpc handler task after an RPC function
    has been executed.
    @param handlerId @b IN Rpc handler Id.
    @param rpcId @b IN Id of Rpc function that has just finished executing.
*/
typedef void phStbRpcTM_HandlerTaskFunc_t(
    phStbRpc_HandlerId_t handlerId,
    UInt32 rpcId
);


/** An entry declaring specifics for an Rpc function within a Rpc handler.
    @see PH_STBRPC_RPCENTRY */
typedef struct phStbRpcTM_RpcEntry {
    /** Id of the function. */
    phStbRpc_FunId_t funId;
    /** Function name for diagnostics. */
    String funcName;
    /** Unmarshal function invoked by the handler for this Rpc function. */
    phStbRpcTM_RpcUnmarshalFunc_t *pUnmarshalFunc;
} phStbRpcTM_RpcEntry_t;


/** Pointer to a ::phStbRpcTM_RpcEntry_t. */
typedef phStbRpcTM_RpcEntry_t *pphStbRpcTM_RpcEntry_t;


/** Information for a Rpc handler on a target CPU.
    Either pGlobalUnmarshalFunc or pRpcEntries should be Non Null.
    @see phStbRpc_Handler_t
        phStbRpcTM_HandlerInit
*/
typedef struct phStbRpcTM_Handler phStbRpcTM_Handler_t;


/** Pointer to a ::phStbRpcTM_Handler_t. */
typedef phStbRpcTM_Handler_t *pphStbRpcTM_Handler_t;


/*******************************************************************
* EXPORTED DATA      g[k|p|kp|pk|kpk]ph[<lnx|tm|NONE>]StbRpcTM_<Word>+ *
********************************************************************/

/*******************************************************************
* EXPORTED FUNCTION PROTOTYPES     ph[<lnx|tm|NONE>]StbRpcTM[_<Word>+]   *
********************************************************************/

/** Initialises this component.
*/
tmErrorCode_t phStbRpcTM_Init(void);


/******************************************************************************/
/** Register a handler for rpc functions. This does not start it yet.
    Either pGlobalUnmarshalFunc or rpcEntries should be non-NULL (not both).
    @param name @b IN Name of the channel (used for debug purposes).
    @param handlerId @b IN Id of the RPC handler on the target.
    @param numFun @b IN Number of Rpc functions registered for this handler.
    @param payloadMaxLen @b IN Maximum RPC payload data length in bytes.
    @param pGlobalUnmarshalFunc @b IN Function to handle all rpc requests.
    Must be NULL if rpcEntries is !NULL.
    @param rpcEntries @b IN Table of entries detailing each rpc function.
    Must be NULL if pGlobalUnmarshalFunc is !NULL.
    @param ppHandler @b OUT The handler.
    @return PH_OK or error code.
*/
tmErrorCode_t phStbRpcTM_HandlerCreate(
    Char *name,
    phStbRpc_HandlerId_t handlerId,
    UInt32 numFun,
    UInt32 payloadMaxLen,
    phStbRpcTM_RpcUnmarshalFunc_t pGlobalUnmarshalFunc,
    phStbRpcTM_RpcEntry_t rpcEntries[],
    pphStbRpcTM_Handler_t *ppHandler
);

/******************************************************************************/
/** Starts running a handler for rpc functions.
    This creates a task to handle the requests.
    @param pHandler @b IN The handler.
    @param priority @b IN Priority (for tmosal) of the task created to handle the Rpc requests.
    @param handlerTaskFunc @b IN Optional function to be called after an Rpc function
    has been handled.
    @return PH_OK or error code.
*/
tmErrorCode_t phStbRpcTM_HandlerStart(
    pphStbRpcTM_Handler_t pHandler,
    tmosalTaskPriority_t priority,
    phStbRpcTM_HandlerTaskFunc_t *handlerTaskFunc
);

/******************************************************************************/
/** Gets a RPC request for an rpc function.
    @param pHandler @b IN The handler.
    @return The request data.
*/
pphStbRpc_RpcBuf_t phStbRpcTM_GetRequest(
    pphStbRpcTM_Handler_t pHandler
);

/******************************************************************************/
/** Posts return results for a rpc function.
    Must always be called, even for a asynchronous RPC call.

    For a synchronous RPC call, the rpc buffer returned by the
    previous ::phStbRpcTM_GetRequest call must have been updated before
    calling phStbRpcTM_Return:
    - payload data must have been stored.

    When this function is called, the data (::phStbRpc_RpcBuf_t)
    returned by the previous phStbRpcTM_GetRequest() call will be automatically reclaimed.

    @param pHandler @b IN The handler.
    @param rpcStatus @b IN Status of the RPC process (not the function itself);
    0 is OK.
    This is not the result of the function itself (see the result param for this).
    @param result @b IN Result of the function.
    @param size @b IN Size of data returned in the RPC buffer obtained from phStbRpcTM_GetRequest().
    Must be 0 for a asynchronous RPC call (see ::phStbRpc_Option_Async).
    Thus the return data is to be stored in the same rpc buffer that was obtained
    by from ::phStbRpcTM_GetRequest.
*/
void phStbRpcTM_Return(
    pphStbRpcTM_Handler_t pHandler,
    UInt32 rpcStatus,
    UInt32 result,
    UInt32 size
);

#endif /* PHSTBRPCTM_H      Do not add any thing below this line */
