#ifndef PHSTBRPC_H    /* Multi-include protection.   <FILENAME>_H */
#define PHSTBRPC_H

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright (C) 2006 NXP B.V.
 * All Rights Reserved.
*//**
    @file   phStbRpc.h
    @brief  Linux Proxy for RPC.

@b  Component: phStbRpc.

*//*

Rev Date        Author      Comments
--------------------------------------------------------------------------------
  1 20050418    batelaan    Original
  2 20050511    batelaan    Update doc; add error codes; add pResult param to phStbRpc_RpcCall()
  3 20050512    batelaan    Added PH_STBRPC_ILLEGAL_RPCID.
  4 20050621    batelaan    Char->UInt8
  5 20050712    batelaan    Add phStbRpc_ProxyDeInit
  6 20050808    batelaan    Added PH_ERR_STBRPC_XFER
  7 20050810    batelaan    Added payloadMaxLen param to phStbRpc_ProxyInit
  8 20050920    burningh    Used CID_STBRPC in error codes
  9 20050920    burningh    Crosscheck changes
 10 20051001    batelaan    Add phStbRpc_Init
 11 20051004    batelaan    Added PH_ERR_STBRPC_INTERRUPTED
 12 20051117    laird       Added phStbRpc_DeInit
 13 20060124    batelaan    Add extern "C" for use with C++
 14 20070730    laird       Updated Header
--------------------------------------------------------------------------------
    For consistency and standardisation retain the Section Separators.
*/


/*******************
* INCLUDE FILES    *
********************/

#include <phStbCommon.h>
#include <phStbRpc_Common.h>

#ifdef __cplusplus
extern "C"
{
#endif

/**************************************************************
* EXPORTED MACROS         PH[<LNX|TM|NONE>]_IPSTBRPC_<_WORD>+        *
***************************************************************/

/* This hard coded value is present because we can't include tmNxCompId.h on Linux */
#define PH_ERR_STBRPC_BASE           0x88226000
#define PH_ERR_STBRPC_UNIQUE_BASE   (PH_ERR_STBRPC_BASE | 0x800)

/** RPC Channel was/is closed error, or could not be opened. */
#define PH_ERR_STBRPC_CLOSED        (PH_ERR_STBRPC_UNIQUE_BASE | 1L)

/** Insufficient memory error. */
#define PH_ERR_STBRPC_NO_MEM        (PH_ERR_STBRPC_UNIQUE_BASE | 2L)

/** RPC Channel is not initialised error. */
#define PH_ERR_STBRPC_NOT_INIT      (PH_ERR_STBRPC_UNIQUE_BASE | 3L)

/** Illegal handler Id error. */
#define PH_ERR_STBRPC_HANDLERID     (PH_ERR_STBRPC_UNIQUE_BASE | 4L)

/** Indicates that an illegal funId has been received. */
#define PH_ERR_STBRPC_FUNID         (PH_ERR_STBRPC_UNIQUE_BASE | 5L)

/** Unspecified low level interprocessor transfer error. Usually fatal. */
#define PH_ERR_STBRPC_XFER          (PH_ERR_STBRPC_UNIQUE_BASE | 6L)

/** RPC call was interrupted. It can be restarted. */
#define PH_ERR_STBRPC_INTERRUPTED   (PH_ERR_STBRPC_UNIQUE_BASE | 7L)

/**************************************************************
* EXPORTED TYPEDEFS       [p]ph[<lnx|tm|NONE>]StbRpc_<Word>+_t  *
***************************************************************/

/** An entry declaring specifics for an Rpc function within a Rpc handler.
    @see PH_STBRPC_RPCENTRY */
typedef struct phStbRpc_RpcProxyEntry {
    /** Options for this rpc function. */
    phStbRpc_Option_t options;
} phStbRpc_RpcProxyEntry_t;

/** Pointer to a ::phStbRpc_RpcProxyEntry_t. */
typedef phStbRpc_RpcProxyEntry_t *pphStbRpc_RpcProxyEntry_t;

/*******************************************************************
* EXPORTED DATA      g[k|p|kp|pk|kpk]ph[<lnx|tm|NONE>]StbRpc_<Word>+ *
********************************************************************/

/*******************************************************************
* EXPORTED FUNCTION PROTOTYPES     ph[<lnx|tm|NONE>]StbRpc[_<Word>+]   *
********************************************************************/

/******************************************************************************/
/** Initialises the RPC API.
    Must be called by an application before using RPC.
    It can be called for example from DllMain on Windows.
*/
void phStbRpc_Init(Void);

/******************************************************************************/
/** De-Initialises the RPC API.
    Must be called by an application before they can shutdown RPC.
*/
void phStbRpc_DeInit(Void);

/******************************************************************************/

/** Register a handler on the host side for rpc functions.
    @param handlerId @b IN RPC handler id.
    @param numRpc @b IN Number of Rpc functions registered for this handler.
    @param payloadMaxLen @b IN Maximum RPC payload data length in bytes.
*/
tmErrorCode_t phStbRpc_ProxyInit(
    phStbRpc_HandlerId_t handlerId,
    UInt32 numRpc,
    UInt32 payloadMaxLen
);

/******************************************************************************/

/** Deregister the handler on the host side for rpc functions.
    @param handlerId @b IN RPC handler id.
*/
tmErrorCode_t phStbRpc_ProxyDeInit(
    phStbRpc_HandlerId_t handlerId);

/******************************************************************************/

/** Call a function via RPC.
    @param handlerId @b IN RPC handler id.
    @param funId @b IN Id of the RPC function, specific to the handler.
    @param flags @b IN Flags for special options. See ::phStbRpc_Option_t.
    @param pBuffer @b INOUT Buffer with data to be transferred to the target. This
    will be updated by the RPC call, unless flags contains ::phStbRpc_Option_Async.
    @param pBufferSize @b INOUT Size of the buffer to be transferred,
    and amount of bytes returned (unless ::phStbRpc_Option_Async is used).
    @param pResult OUT Pointer to return value from the function itself. May be NULL
    for an ::phStbRpc_Option_Async function call.
    @return
    - 0 if no RPC error occured.
    - Otherwise error code; see PH_STBRPC_* codes.
*/
tmErrorCode_t phStbRpc_RpcCall(
    phStbRpc_HandlerId_t handlerId,
    UInt32 funId,
    UInt32 flags,
    pUInt8 pBuffer,
    pUInt32 pBufferSize,
    pUInt32 pResult);

/******************************************************************************/

#ifdef __cplusplus
}
#endif

#endif /* PHSTBRPC_H      Do not add any thing below this line */
