#ifndef PHSTBRPC_XFER_H    /* Multi-include protection.   <FILENAME>_H */
#define PHSTBRPC_XFER_H

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright (C) 2006 NXP B.V.
 * All Rights Reserved.
*//**
    @file   phStbRpc_Xfer.h
    @brief  Interprocessor data transfer interface for RPC.

    This file specifies an interface to a data tranfer API used by this module to transfer
    RPC requests and responses. This interface is used to abstract away the actual low level transfer
    mechanism (e.g. Luddite or tmman).

    The interpretation of some of the parameters to the phStbRpc_Xfer_Receive() function differ
    slightly depending on whether this interface is used on the proxy side (e.g. Linux/WinCE)
    or on the RPC server side (target, e.g. TriMedia).

    On the server side when calling phStbRpc_Xfer_Receive() the function must create and
    return a buffer with the RPC data. On the proxy side however this function re-uses the
    buffer that was originally used to send the RPC request, and thus no buffer allocation is needed.

@b  Component: phStbRpc.

*//*

Rev Date        Author      Comments
--------------------------------------------------------------------------------
  1 20050418    batelaan    Original
  2 20050506    batelaan    Corrected PH_STBRPC_DEVICE_NAME_PREFIX. Update api.
  3 20050513    batelaan    Crosscheck fixes.
  4 20050621    batelaan    Char->UInt8
  5 20050712    batelaan    Add phStbRpc_ProxyDeInit
  6 20050808    batelaan    Port to tmman.
  7 20050808    batelaan    Added phStbRpc_Xfer_Init().
  8 20050810    batelaan    Added payloadMaxLen param to phStbRpc_XferCreate
  9 20050906    batelaan    Enable second TM.
 10 20051121    laird       Adding De-Init function.
 11 20051125    burningh    Adding TaskCreate function
 12 20060124    batelaan    Add extern "C" for use with C++
 13 20070730    laird       Updated Header
--------------------------------------------------------------------------------
    For consistency and standardisation retain the Section Separators.
*/


/*******************
* INCLUDE FILES    *
********************/

#include <phStbCommon.h>
#include <phStbRpc_Common.h>

#ifdef __cplusplus
extern "C"
{
#endif

/**************************************************************
* EXPORTED MACROS         PH[<LNX|TM|NONE>]_IPSTBRPC_<_WORD>+        *
***************************************************************/

/**************************************************************
* EXPORTED TYPEDEFS       [p]ph[<lnx|tm|NONE>]StbRpc_<Word>+_t  *
***************************************************************/

/** Used for transferring data to/from target CPU.
    There are host and target implementations of this. */
typedef struct phStbRpc_Xfer phStbRpc_Xfer_t;

/** Pointer to a ::phStbRpc_Xfer_t. */
typedef phStbRpc_Xfer_t *pphStbRpc_Xfer_t;

/*******************************************************************
* EXPORTED DATA      g[k|p|kp|pk|kpk]ph[<lnx|tm|NONE>]StbRpc_<Word>+ *
********************************************************************/

/*******************************************************************
* EXPORTED FUNCTION PROTOTYPES     ph[<lnx|tm|NONE>]StbRpc[_<Word>+]   *
********************************************************************/

/******************************************************************************/

/** Initialise this interface.
*/
void phStbRpc_Xfer_Init(void);

/******************************************************************************/

/** De-Initialise this interface.
*/
void phStbRpc_Xfer_DeInit(void);

/******************************************************************************/

/** Create a new transfer channel for RPC.
    @param targetId @b IN RPC target CPU id.
    @param handlerId @b IN RPC handler id.
    @param maxBufferSize @b IN Maximum amount of bytes that a single transfer
    must be able to send. This may be ignored if the implementation of this
    interface has no size restrictions.
    @param ppXfer @b OUT Transfer handle.
    @return TM_OK if successful.
*/
tmErrorCode_t phStbRpc_Xfer_Create(
    phStbRpc_TargetId_t targetId,
    phStbRpc_HandlerId_t handlerId,
    UInt32 maxBufferSize,
    pphStbRpc_Xfer_t *ppXfer);

/******************************************************************************/

/** Destroy a  transfer channel for RPC.
    @param pXfer @b IN Transfer handle.
*/
Void phStbRpc_Xfer_Destroy(
    pphStbRpc_Xfer_t pXfer);

/******************************************************************************/

/** Send a RPC request or response.
    @param pXfer @b IN Transfer handle.
    @param pBuffer @b IN Buffer with RPC data to send.
    @param bufferSize @b IN Size of the buffer to be sent.
    If sending an RPC response, then if this is zero it indicates an
    asynchronous call, and nothing should be sent back (if possible);
    however, the buffer should be recycled for reuse.
    @return
    - 0 if OK
    - otherwise an error code.
*/
tmErrorCode_t phStbRpc_Xfer_Send(
    pphStbRpc_Xfer_t pXfer,
    pUInt8 pBuffer,
    UInt32 bufferSize);

/******************************************************************************/

/** Receive a RPC request or response.
    If receiving a RPC response, then for an asynchronous call no data is read
    (*pRpcBufSize must be 0).
    @param pXfer @b IN Transfer handle.
    @param ppBuffer @b INOUT Buffer to be filled.
    For receiving a RPC response the buffer already exists.
    For receiving a RPC request the buffer must be created and returned via this parameter.
    @param maxRpcBufSize @b IN Maximum size of the buffer (only applicable
    for receiving a RPC response.)
    @return
    - -1 if the RPC channel has been closed
    - Otherwise amount of bytes read (may be 0).
*/
Int32 phStbRpc_Xfer_Receive(
    pphStbRpc_Xfer_t pXfer,
    pUInt8 *ppBuffer,
    UInt32 maxRpcBufSize);

/******************************************************************************/

/** Allocate memory for RPC.
    @param pXfer @b IN Transfer handle.
    @param size @b IN Size to be allocated in bytes.
    @return Address.
*/
void * phStbRpc_Xfer_Malloc(
    pphStbRpc_Xfer_t pXfer,
    UInt32 size);

/******************************************************************************/

/** Free memory for RPC.
    @param pXfer @b IN Transfer handle.
    @param address @b IN Address to be freed.
*/
void phStbRpc_Xfer_Free(
    pphStbRpc_Xfer_t pXfer,
    void *address);

/******************************************************************************/

/** Create a task in an Xfer-implementation defined manner.
    @param pTaskName @b IN The name of the task.
    @param handlerFunction @b IN The function that is executed in the newly created task.
    @param taskPriority @b IN The priority of the newly created task.
    @param pTaskArg @b IN The argument that is passed to the handlerFunction.
    @return
    - 0 if OK
    - otherwise an error code.
*/
tmErrorCode_t phStbRpc_Xfer_TaskCreate(
    char *pTaskName,
    void (*handlerFunction)(void *),
    UInt32 taskPriority,
    void *pTaskArg);

#ifdef __cplusplus
}
#endif

#endif /* PHSTBRPC_XFER_H      Do not add any thing below this line */
